import os
import subprocess as sp
import warnings
from typing import List

__all__ = ['encode_video_comparison']


def encode_video_comparison(
        audiofile: str,
        subtitle_files: List[str],
        output_videopath: str = None,
        *,
        labels: List[str] = None,
        height: int = 90,
        width: int = 720,
        color: str = 'black',
        fontsize: int = 70,
        border_color: str = 'white',
        label_color: str = 'white',
        label_size: int = 14,
        fps: int = 25,
        video_codec: str = None,
        audio_codec: str = None,
        overwrite=False,
        only_cmd: bool = False,
        verbose=True
) -> (str, None):
    """
    Encode multiple subtitle files into one video with the subtitles vertically stacked.

    Parameters
    ----------
    audiofile : str
        Path of audio file.
    subtitle_files : list of str
        List of paths for subtitle file.
    output_videopath : str, optional
        Output video path.
    labels : list of str, default, None, meaning use ``subtitle_files`` as labels
        List of labels for ``subtitle_files``.
    height : int, default 90
        Height for each subtitle section.
    width : int, default 720
        Width for each subtitle section.
    color : str, default 'black'
        Background color of the video.
    fontsize: int, default 70
        Font size for subtitles.
    border_color : str, default 'white'
        Border color for separating the sections of subtitle.
    label_color : str, default 'white'
        Color of labels.
    label_size : int, default 14
        Font size of labels.
    fps : int, default 25
        Frame-rate of the video.
    video_codec : str, optional
        Video codec opf the video.
    audio_codec : str, optional
        Audio codec opf the video.
    overwrite : bool, default False
        Whether to overwrite existing video files with the same path as the output video.
    only_cmd : bool, default False
        Whether to skip encoding and only return the full command generate from the specified options.
    verbose : bool, default True
        Whether to display ffmpeg processing info.

    Returns
    -------
    str or None
        Encoding command as a string if ``only_cmd = True``.
    """
    vc = '' if video_codec is None else f' -c:v {video_codec}'
    ac = '' if audio_codec is None else f' -c:a {audio_codec}'
    background = f'-f lavfi -i color=size={width}x{height}:rate={fps}:color={color}'
    border = f'-f lavfi -i color=size={width}x3:rate={fps}:color={border_color}'
    audio = f'-i "{audiofile}"'
    cfilters0 = []
    assert labels is None or len(labels) == len(subtitle_files)
    for i, sub in enumerate(subtitle_files):
        label = sub if labels is None else labels[i]
        label = label.replace("'", '"')
        fil = f"[0]drawtext=text='{label}':fontcolor={label_color}:fontsize={label_size}:x=10:y=10[a{i}]," \
              f"[a{i}]subtitles='{sub}':force_style='Fontsize={fontsize}'[b{i}]"
        cfilters0.append(fil)
    cfilters1 = (
            '[1]'.join(
                f'[b{i}]' for i in range(len(cfilters0))
            )
            +
            f'vstack=inputs={len(cfilters0) * 2 - 1}'
    )
    final_fil = ','.join(cfilters0) + f';{cfilters1}'
    ow = '-y' if overwrite else '-n'
    if output_videopath is None:
        name = os.path.split(os.path.splitext(audiofile)[0])[1]
        output_videopath = f'{name}_sub_comparison.mp4'
    cmd = (f'ffmpeg {ow} {background} {border} {audio} '
           f'-filter_complex "{final_fil}"{vc}{ac} -shortest "{output_videopath}"')
    if only_cmd:
        return cmd
    if verbose:
        print(cmd)
    rc = sp.run(cmd, capture_output=not verbose).returncode
    if rc == 0:
        if verbose:
            print(f'Encoded: {output_videopath}')
    else:
        warnings.warn(f'Failed to encode {output_videopath}')




import subprocess
import warnings
import torch
import numpy as np
from typing import Union, Optional, Tuple

from .utils import (
    is_ytdlp_available, load_source, load_audio, voice_freq_filter, get_samplerate, get_metadata
)
from .demucs import is_demucs_available, load_demucs_model, demucs_audio
from .dfnet import is_dfnet_available, load_dfnet_model, dfnet_audio
from .noisereduce import is_noisereduce_available, load_noisereduce_model, noisereduce_audio
from .output import save_audio_tensor
from ..utils import update_options

from ..whisper_compatibility import SAMPLE_RATE


SUPPORTED_DENOISERS = {
    'demucs': {'run': demucs_audio, 'load': load_demucs_model, 'access': is_demucs_available},
    'dfnet': {'run': dfnet_audio, 'load': load_dfnet_model, 'access': is_dfnet_available},
    'noisereduce': {'run': noisereduce_audio, 'load': load_noisereduce_model, 'access': is_noisereduce_available}
}


def get_denoiser_func(denoiser: Optional[str], key: str):
    if denoiser is None:
        return
    if denoiser not in SUPPORTED_DENOISERS:
        raise NotImplementedError(f'"{denoiser}" is one of the supported denoisers: '
                                  f'{tuple(SUPPORTED_DENOISERS.keys())}.')
    assert key in ('run', 'load', 'access')
    return SUPPORTED_DENOISERS[denoiser][key]


def _load_file(**kwargs):
    warnings.warn('This function is deprecated. Use `stable_whisper.audio.load_source()`.',
                  stacklevel=2)
    return load_source(**kwargs)


def convert_demucs_kwargs(
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: bool = None,
        demucs_options: Optional[dict] = None,
) -> Tuple[Union[str, None], dict]:
    if demucs:
        warnings.warn('``demucs`` is deprecated and will be removed in future versions. '
                      'Use ``denoiser="demucs"`` instead.',
                      stacklevel=2)
        if denoiser:
            if denoiser != 'demucs':
                raise ValueError(f'Demucs is enabled but got "{denoiser}" for denoiser.')
        else:
            denoiser = 'demucs'

    if denoiser_options is None:
        denoiser_options = {}

    if demucs_options:
        warnings.warn('``demucs_options`` is deprecated and will be removed in future versions. '
                      'Use ``denoiser_options`` instead.',
                      stacklevel=2)
        if denoiser == 'demucs':
            if demucs_options:
                denoiser_options = demucs_options
            if isinstance(demucs, torch.nn.Module):
                denoiser_options['model'] = demucs

    return denoiser, denoiser_options


def prep_audio(
        audio: Union[str, np.ndarray, torch.Tensor, bytes],
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        only_voice_freq: bool = False,
        only_ffmpeg: bool = False,
        verbose: Optional[bool] = False,
        sr: int = None,
        demucs: Optional[str] = None,
        demucs_options: Optional[dict] = None,
) -> torch.Tensor:
    """
    Converts input audio of many types into a mono waveform as a torch.Tensor.

    Parameters
    ----------
    audio : str or numpy.ndarray or torch.Tensor or bytes
        Path/URL to the audio file, the audio waveform, or bytes of audio file.
        If audio is :class:`numpy.ndarray` or :class:`torch.Tensor`, the audio must be already at sampled to 16kHz.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    sr : int, default None, meaning ``whisper.audio.SAMPLE_RATE``, 16kHZ
        The sample rate of ``audio``.
    verbose : bool or None, default True
        Verbosity for yd-dlp and displaying content metadata when ``file`` is a URL. If not ``None``, display metadata.
        For yd-dlp: ``None`` is "--quiet"; ``True`` is "--progress"; ``False`` is "--progress" + "--quiet".
    only_ffmpeg: bool, default False
        Whether to use only FFmpeg (and not yt-dlp) for URls.

    Returns
    -------
    torch.Tensor
        A mono waveform.
    """
    denoiser, denoiser_options = convert_demucs_kwargs(
        denoiser, denoiser_options, demucs=demucs, demucs_options=demucs_options
    )
    if not sr:
        sr = SAMPLE_RATE

    denoise_func = get_denoiser_func(denoiser, 'run')

    if isinstance(audio, (str, bytes)):
        if denoise_func is None:
            audio = torch.from_numpy(load_audio(audio, sr=sr, verbose=verbose, only_ffmpeg=only_ffmpeg))
        else:
            denoiser_options = update_options(
                denoiser_options,
                True,
                audio=audio,
                output_sr=sr,
                verbose=verbose
            )
            audio = denoise_func(**denoiser_options)
    else:
        if isinstance(audio, np.ndarray):
            audio = torch.from_numpy(audio)
        if denoise_func is not None:
            denoiser_options = update_options(
                denoiser_options,
                True,
                audio=audio,
                input_sr=sr,
                output_sr=sr,
                verbose=verbose,
            )
            audio = denoise_func(**denoiser_options)
    if only_voice_freq:
        audio = voice_freq_filter(audio.cpu(), sr)

    return audio


class AudioLoader:
    def __init__(
            self,
            source: Union[str, np.ndarray, torch.Tensor, bytes],
            buffer_size: Union[int, str] = None,
            stream: Optional[bool] = None,
            sr: int = None,
            test_first_chunk: bool = True,
            verbose: Optional[bool] = False,
            only_ffmpeg: bool = False,
            new_chunk_divisor: Optional[int] = 512,
            save_path: Optional[str] = None,
            post_prep_callback=None,
            denoiser: Optional[str] = None,
            denoiser_options: Optional[dict] = None,
            only_voice_freq: bool = False,
            demucs: Optional[str] = None,
            demucs_options: Optional[dict] = None,
    ):
        if stream and not isinstance(source, str):
            raise NotImplementedError(f'``stream=True`` only supported for string ``source`` but got {type(source)}.')
        self.source = source
        if sr is None:
            from whisper.audio import SAMPLE_RATE
            sr = SAMPLE_RATE
        self._sr = sr
        if buffer_size is None:
            buffer_size = (sr * 30)
        self._buffer_size = self._valid_buffer_size(self.parse_chunk_size(buffer_size))
        self._stream = isinstance(source, str) if stream is None else stream
        self._accum_samples = 0
        self.verbose = verbose
        self.only_ffmpeg = only_ffmpeg
        self.new_chunk_divisor = new_chunk_divisor
        self._post_prep_callback = post_prep_callback
        self._denoiser = denoiser
        self._denoiser_options = denoiser_options or {}
        self._denoiser, self._denoiser_options = convert_demucs_kwargs(
            self._denoiser, self._denoiser_options, demucs=demucs, demucs_options=demucs_options
        )
        self._final_save_path = save_path
        self._denoised_save_path = self._denoiser_options.pop('save_path', None)
        self._only_voice_freq = only_voice_freq
        self._denoised_samples_to_save = []
        self._final_samples_to_save = []
        metadata = get_metadata(source)
        self._source_sr, self._duration_estimation = metadata['sr'] or 0, metadata['duration'] or 0
        self._total_sample_estimation = round(self._duration_estimation * self._sr)
        self._denoise_model, self._min_chunk = self._load_denoise_model()
        self.check_min_chunk_requirement()
        self._prep = self._get_prep_func()
        self._extra_process = None
        self._prev_seek = None
        self._buffered_samples = torch.tensor([])
        self._prev_unprep_samples = np.array([])
        self._process = self._audio_loading_process()
        if test_first_chunk and self.next_chunk(0) is None:
            raise RuntimeError(f'FFmpeg failed to read "{source}".')

    @property
    def buffer_size(self):
        return self._buffer_size

    @property
    def sr(self):
        return self._sr

    @property
    def source_sr(self):
        return self._source_sr

    @property
    def stream(self):
        return self._stream

    @property
    def prev_seek(self):
        return self._prev_seek

    @buffer_size.setter
    def buffer_size(self, size: int):
        self._buffer_size = self._valid_buffer_size(size)
        self.check_min_chunk_requirement()

    @staticmethod
    def _valid_buffer_size(size: int):
        if size < 0:
            raise ValueError('buffer size must be at least 0')
        return size

    def parse_chunk_size(self, chunk_size: Union[int, str]) -> int:
        if isinstance(chunk_size, int):
            return chunk_size
        if not chunk_size.endswith('s'):
            raise ValueError('string ``chunk_size`` must end with "s"')
        chunk_size = chunk_size[:-1]
        seconds = float(chunk_size)
        return round(seconds * self._sr)

    def _load_denoise_model(self):
        if not self._denoiser:
            return None, None
        model = get_denoiser_func(self._denoiser, 'load')(True)
        length = int(getattr(model, 'segment', 5) * self._sr)
        return model, length

    def check_min_chunk_requirement(self):
        if self._min_chunk is not None and self._min_chunk > self.buffer_size:
            warnings.warn(
                f'It is advised to keep ``chunk_size`` ({self.buffer_size}) at least {self._min_chunk}'
                f' or "{round(self._min_chunk / self._sr)}s" for ``demucs=True``.',
                stacklevel=2
            )

    def get_duration(self, ndigits: int = None):
        if self._stream:
            dur = (
                self._duration_estimation
                if (dur := (self._accum_samples or 0) / self._sr) < self._duration_estimation else
                dur
            )
        else:
            dur = self._duration_estimation
        return dur if ndigits is None else round(dur, ndigits=ndigits)

    def get_total_samples(self):
        if not self._stream:
            return self._total_sample_estimation
        return (
            self._total_sample_estimation
            if (self._accum_samples / self._sr) < self._duration_estimation else
            self._accum_samples
        )

    def update_post_prep_callback(self, callback):
        self._post_prep_callback = callback
        if self._post_prep_callback is not None and len(self._buffered_samples):
            self._post_prep_callback(self._buffered_samples)

    def __del__(self):
        self.terminate()

    def divisible_min_chunk(self, min_chunk: int) -> int:
        if not self.new_chunk_divisor:
            return min_chunk
        if r := min_chunk % self.new_chunk_divisor:
            return min_chunk + self.new_chunk_divisor - r
        return min_chunk

    def _seek_buffered_samples(self, seek) -> int:
        is_first_time_loading = self._prev_seek is None
        if is_first_time_loading:
            if self._process is None:
                samples_to_load_discard = 0
                self._buffered_samples = self._prep(self.source)
                if self._final_save_path:
                    self._final_samples_to_save.append(self._buffered_samples.cpu())
                self._total_sample_estimation = self._buffered_samples.shape[-1]
                self._duration_estimation = (self._total_sample_estimation / self._sr)
                self._buffered_samples = self._buffered_samples[seek:]
            else:
                samples_to_load_discard = seek
                self._buffered_samples = torch.tensor([])
        else:
            assert seek >= self._prev_seek, '``seek`` must be >= the previous ``seek`` value'
            seek_delta = seek - self._prev_seek
            samples_to_load_discard = max(0, seek_delta - len(self._buffered_samples))
            self._buffered_samples = self._buffered_samples[seek_delta:]

        self._prev_seek = seek

        return samples_to_load_discard

    def _read_samples(self, samples: int) -> bytes:
        if self._process is None or self._process.poll() is not None:
            return b''
        n = samples * 2
        b = b''
        while len(b) < n and self._process.poll() is None:
            nb = self._process.stdout.read(n)
            b += nb
        return b

    def _read_append_to_buffer(self, samples_to_read: int, samples_to_discard: Optional[int] = None):
        sample_bytes = self._read_samples(samples_to_read)
        if not sample_bytes:
            return
        new_samples = self._prep_samples(sample_bytes, samples_to_discard)
        if len(self._buffered_samples):
            self._buffered_samples = torch.concat([self._buffered_samples, new_samples], dim=-1)
        else:
            self._buffered_samples = new_samples

    def save_denoised_audio(self, path: Optional[str] = None):
        if not self._denoised_samples_to_save:
            warnings.warn('Failed to save denoised audio. No stored denoised audio samples found.',
                          stacklevel=2)
            return
        if not (path or self._denoised_save_path):
            warnings.warn('Failed to save denoised audio. No specified path to save.',
                          stacklevel=2)
            return
        save_audio_tensor(torch.cat(self._denoised_samples_to_save), path or self._denoised_save_path, self._sr)

    def save_final_audio(self, path: Optional[str] = None):
        if not self._final_samples_to_save:
            warnings.warn('Failed to save final audio. No stored final audio samples found.',
                          stacklevel=2)
            return
        if not (path or self._final_save_path):
            warnings.warn('Failed to save denoised audio. No specified path to save.',
                          stacklevel=2)
            return
        save_audio_tensor(torch.cat(self._final_samples_to_save), path or self._final_save_path, self._sr)

    def next_chunk(self, seek: int, size: Optional[int] = None) -> Union[torch.Tensor, None]:
        samples_to_load_discard = self._seek_buffered_samples(seek)
        samples_to_load_keep = max(self._buffer_size, size or 0) - len(self._buffered_samples)
        if samples_to_load_keep > 0:
            samples_to_load_keep = self.divisible_min_chunk(samples_to_load_keep)
        samples_to_load = max(samples_to_load_discard + samples_to_load_keep, 0)
        self._read_append_to_buffer(samples_to_load, samples_to_load_discard)

        samples = self._buffered_samples[:self._buffer_size if size is None else size]

        return samples if len(samples) else None

    def _get_prep_func(self):

        if self._denoiser:
            self._denoiser_options['model'] = self._denoise_model
            if 'progress' not in self._denoiser_options:
                self._denoiser_options['progress'] = False if self._stream else (self.verbose is not None)

        if self._stream:
            if self._denoised_save_path:
                if self._final_save_path:
                    warnings.warn('Both ``save_path`` in AudioLoad and ``denoiser_options`` were specified, '
                                  'but only the final audio will be saved for ``stream=True`` in either case. '
                                  '``denoiser_options`` will be prioritized for ``save_path``.',
                                  stacklevel=2)
                else:
                    self._final_save_path = self._denoised_save_path
                self._denoised_save_path = None

            denoise_func = get_denoiser_func(self._denoiser, 'run')

            def prep(audio):
                audio = torch.from_numpy(audio)
                if denoise_func is not None:
                    denoiser_options = update_options(
                        self._denoiser_options,
                        True,
                        audio=audio,
                        input_sr=self._sr,
                        output_sr=self._sr,
                        verbose=self.verbose,
                    )
                    audio = denoise_func(**denoiser_options)
                if self._only_voice_freq:
                    audio = voice_freq_filter(audio.cpu(), self._sr)

                return audio

            return prep

        if self._denoised_save_path:
            def append_denoised(samples: torch.Tensor):
                self._denoised_samples_to_save.append(samples.cpu())

            self._denoiser_options['save_path'] = append_denoised

        def prep(audio):
            return prep_audio(
                audio,
                denoiser=self._denoiser,
                denoiser_options=self._denoiser_options,
                only_voice_freq=self._only_voice_freq,
                only_ffmpeg=self.only_ffmpeg,
                verbose=self.verbose,
                sr=self._sr,
            )

        return prep

    def _prep_samples(self, new_samples: bytes, samples_to_discard: Optional[int] = None) -> torch.Tensor:
        if samples_to_discard:
            assert not len(self._buffered_samples)
            i = samples_to_discard * 2
            discarded_samples_bytes, new_samples = new_samples[:i], new_samples[i:]
        else:
            discarded_samples_bytes = b''
        new_samples = np.frombuffer(new_samples, np.int16).flatten().astype(np.float32) / 32768.0
        new_sample_length = new_samples.shape[-1]
        self._accum_samples += new_sample_length
        if self._min_chunk:
            if (
                    len(self._prev_unprep_samples) + len(discarded_samples_bytes)
                    and (missing_length := self._min_chunk - new_sample_length) > 0
            ):
                prev_unprep_samples = self._prev_unprep_samples
                if discarded_samples_bytes:
                    discarded_samples_bytes = discarded_samples_bytes[-missing_length*2:]
                    discarded_samples_bytes = np.frombuffer(
                        discarded_samples_bytes, np.int16
                    ).flatten().astype(np.float32) / 32768.0
                    if discarded_samples_bytes.shape[-1] < missing_length:
                        prev_unprep_samples = np.concatenate((prev_unprep_samples, discarded_samples_bytes), axis=-1)
                    else:
                        prev_unprep_samples = discarded_samples_bytes
                prev_unprep_samples = prev_unprep_samples[-missing_length:]
                new_samples = np.concatenate((prev_unprep_samples, new_samples), axis=-1)
                prepped_samples = self._prep(new_samples)[-new_sample_length:]
            else:
                prepped_samples = self._prep(new_samples)
            self._prev_unprep_samples = new_samples
        else:
            prepped_samples = self._prep(new_samples)

        if self._final_save_path:
            self._final_samples_to_save.append(prepped_samples.cpu())
        if self._post_prep_callback is not None:
            self._post_prep_callback(prepped_samples)

        return prepped_samples

    def terminate(self):
        if getattr(self, '_extra_process', None) is not None and self._extra_process.poll() is None:
            self._extra_process.terminate()
        if getattr(self, '_process', None) is not None and self._process.poll() is None:
            self._process.terminate()
        if getattr(self, '_denoised_save_path'):
            self.save_denoised_audio()
        if getattr(self, '_final_save_path', None):
            self.save_final_audio()

    def _audio_loading_process(self):
        if not isinstance(self.source, str) or not self._stream:
            return
        only_ffmpeg = False
        source = load_source(self.source, verbose=self.verbose, only_ffmpeg=only_ffmpeg, return_dict=True)
        if isinstance(source, dict):
            info = source
            source = info.pop('popen')
        else:
            info = None
        if info and info['duration']:
            self._duration_estimation = info['duration']
            if not self._stream and info['is_live']:
                warnings.warn('The audio appears to be a continuous stream but setting was set to `stream=False`.')

        if isinstance(source, subprocess.Popen):
            self._extra_process, stdin = source, source.stdout
        else:
            stdin = None
        try:
            # This launches a subprocess to decode audio while down-mixing and resampling as necessary.
            # Requires the ffmpeg CLI in PATH.
            cmd = [
                "ffmpeg",
                "-loglevel", "error",
                "-nostdin",
                "-threads", "0",
                "-i", self.source if stdin is None else "pipe:",
                "-f", "s16le",
                "-ac", "1",
                "-acodec", "pcm_s16le",
                "-ar", str(self._sr),
                "-"
            ]
            out = subprocess.Popen(cmd, stdin=stdin, stdout=subprocess.PIPE)

        except subprocess.SubprocessError as e:
            raise RuntimeError(f"Failed to load audio: {e}") from e

        return out

    def validate_external_args(
            self,
            sr=None,
            vad=None,
            stream=None,
            denoiser=None,
            denoiser_options=None,
            only_voice_freq=False
    ):
        if sr and sr != self._sr:
            raise ValueError(
                f'AudioLoader must be initialized with ``sr={sr}`` but ``sr`` of this instance is {self._sr}.'
            )
        if vad:
            from ..stabilization.silero_vad import VAD_WINDOWS
            if self.new_chunk_divisor not in VAD_WINDOWS:
                raise ValueError(
                    f'When ``vad=True``, AudioLoader must be initialized with ``chunk_divisor`` '
                    f'set to one of the following values: {VAD_WINDOWS}. '
                    f'But got {self.new_chunk_divisor} instead.'
                )
        if stream and not self._stream:
            warnings.warn(
                f'``stream=True`` will have no effect unless specified at AudioLoader initialization.',
                stacklevel=2
            )

        if denoiser and not self._denoiser:
            warnings.warn(
                '``denoiser`` will have no effect unless specified at AudioLoader initialization.',
                stacklevel=2
            )

        if denoiser_options and denoiser_options != self._denoiser_options:
            warnings.warn(
                f'``denoiser_options`` does not match the ``denoiser_options`` '
                f'AudioLoader was initialized with.',
                stacklevel=2
            )

        if only_voice_freq and not self._only_voice_freq:
            warnings.warn(
                '``only_voice_freq=True`` will have no effect unless specified at AudioLoader initialization.',
                stacklevel=2
            )
        return self._stream, self._denoiser, self._denoiser_options, self._only_voice_freq


def audioloader_not_supported(audio):
    if isinstance(audio, AudioLoader):
        raise NotImplementedError('This function does not support AudioLoader instances.')



import warnings
from typing import Optional, List

import torch
from tqdm import tqdm

from .utils import SetTorchThread
from ..default import cached_model_instances

VAD_SAMPLE_RATES = (16000, 8000)
VAD_WINDOWS = (256, 512, 768, 1024, 1536)


def load_silero_vad_model(onnx=False, verbose: Optional[bool] = False, cache: bool = True):
    model_cache = cached_model_instances['silero_vad'] if cache else None
    if model_cache is not None and model_cache[onnx] is not None:
        return model_cache[onnx]

    model, utils = torch.hub.load(repo_or_dir='snakers4/silero-vad:master',
                                  model='silero_vad',
                                  verbose=verbose,
                                  onnx=onnx,
                                  trust_repo=True)
    get_ts = utils[0]
    if model_cache is not None:
        model_cache[onnx] = (model, get_ts)
    warnings.filterwarnings('ignore', message=r'operator \(\) profile_node.*', category=UserWarning)

    return model, get_ts


def compute_vad_probs(
        model,
        audio: torch.Tensor,
        sampling_rate: int,
        window: int,
        progress: bool = True
) -> List[float]:
    duration = round(audio.shape[-1] / sampling_rate, 2)
    speech_probs = []
    with SetTorchThread(1), tqdm(total=duration, unit='sec', desc='VAD', disable=not progress) as pbar:
        for current_start_sample in range(0, audio.shape[-1], window):
            chunk = audio[current_start_sample: current_start_sample + window]
            if len(chunk) < window:
                chunk = torch.nn.functional.pad(chunk, (0, int(window - len(chunk))))
            prob = model(chunk.cpu(), sampling_rate).item()
            speech_probs.append(prob)
            if not pbar.disable:
                seek_duration = min(
                    round((current_start_sample + window) / sampling_rate, 2),
                    duration
                )
                pbar.update(seek_duration - pbar.n)

    return speech_probs


def assert_sr_window(sr: int, window: int):
    assert sr in VAD_SAMPLE_RATES, f'{sr} not in {VAD_SAMPLE_RATES}'
    assert window in VAD_WINDOWS, f'{window} not in {VAD_WINDOWS}'



import warnings
from typing import Union, Optional, List

import numpy as np
import torch
import torchaudio

from ..default import is_allow_overwrite


def save_audio_tensor(
        audio: Union[np.ndarray, torch.Tensor],
        path: str,
        sr: int,
        verbose: Optional[bool] = False,
        silent_timings: Optional[Union[np.ndarray, List[dict]]] = None,
        channel: Optional[str] = 'l',
        overwrite: Optional[bool] = None
):
    """
    Save ``audio`` to ``path`` with sections muted according to ``silent_timing`` on ``channel``.
    """
    if not is_allow_overwrite(path, overwrite):
        return
    if channel is not None and channel not in ('l', 'r'):
        raise ValueError(f'``split`` must be "l" or "r" but got "{channel}".')
    if isinstance(audio, np.ndarray):
        audio = torch.from_numpy(audio)
    else:
        audio = audio.cpu()
    dims = audio.ndim
    if dims == 3 and audio.shape[0] == 1:
        audio = audio[0]
    if 0 == dims or dims > 2:
        warnings.warn(f'{dims}D audio Tensor not supported.')
        return
    if dims == 1:
        audio = audio[None]

    if silent_timings is not None:
        audio = audio.mean(dim=0) if audio.shape[0] == 2 else audio[0]
        audio_copy = audio.clone()
        for t in silent_timings:
            s, e = (t['start'], t['end']) if isinstance(t, dict) else t
            s = round(s * sr)
            e = round(e * sr)
            audio_copy[s:e] = 0
        if channel is None:
            audio = audio_copy[None]
        else:
            audio = (audio_copy, audio) if channel == 'l' else (audio, audio_copy)
            audio = torch.stack(audio, dim=0)

    try:
        torchaudio.save(path, audio, sr)
    except ValueError as e:
        warnings.warn(f'Failed to save audio to "{path}". Error: {e}', stacklevel=2)
    else:
        if verbose is not None:
            print(f'Saved: "{path}"')



import warnings
import importlib.metadata
from importlib.util import find_spec
from .utils import get_func_parameters, exact_div

IS_WHISPERLESS_VERSION = True
IS_WHISPER_AVAILABLE = find_spec('whisper') is not None


_COMPATIBLE_WHISPER_VERSIONS = (
    '20230314',
    '20230918',
    '20231105',
    '20231106',
    '20231117',
)
_required_whisper_ver = _COMPATIBLE_WHISPER_VERSIONS[-1]

if IS_WHISPER_AVAILABLE:
    import whisper.tokenizer
    _TOKENIZER_PARAMS = get_func_parameters(whisper.tokenizer.get_tokenizer)
else:
    _TOKENIZER_PARAMS = ()


def whisper_not_available(*args, **kwargs):
    raise ModuleNotFoundError("Please install Whisper: "
                              "'pip install openai-whisper==20231117'. "
                              "Official Whisper repo: https://github.com/openai/whisper")


class Unavailable:
    __init__ = whisper_not_available


if IS_WHISPER_AVAILABLE:
    import whisper
    from whisper.audio import (
        SAMPLE_RATE, N_FRAMES, HOP_LENGTH, N_SAMPLES_PER_TOKEN, N_SAMPLES,
        TOKENS_PER_SECOND, N_FFT, FRAMES_PER_SECOND, CHUNK_LENGTH
    )
    from whisper.tokenizer import LANGUAGES, TO_LANGUAGE_CODE

    from whisper.audio import log_mel_spectrogram, pad_or_trim
    from whisper.timing import median_filter, dtw, merge_punctuations
    from whisper.tokenizer import get_tokenizer as get_whisper_tokenizer

    from whisper.tokenizer import Tokenizer
    from whisper.model import Whisper
    from whisper.decoding import DecodingTask, DecodingOptions, DecodingResult, SuppressTokens
else:
    import torch
    import numpy as np
    whisper = None
    # hard-coded audio hyperparameters from Whisper
    SAMPLE_RATE = 16000
    N_FFT = 400
    HOP_LENGTH = 160
    CHUNK_LENGTH = 30
    N_SAMPLES = CHUNK_LENGTH * SAMPLE_RATE  # 480000 samples in a 30-second chunk
    N_FRAMES = exact_div(N_SAMPLES, HOP_LENGTH)  # 3000 frames in a mel spectrogram input
    N_SAMPLES_PER_TOKEN = HOP_LENGTH * 2  # the initial convolutions has stride 2
    FRAMES_PER_SECOND = exact_div(SAMPLE_RATE, HOP_LENGTH)  # 10ms per audio frame
    TOKENS_PER_SECOND = exact_div(SAMPLE_RATE, N_SAMPLES_PER_TOKEN)  # 20ms per audio token

    log_mel_spectrogram = median_filter = dtw = merge_punctuations = get_whisper_tokenizer \
        = whisper_not_available
    Tokenizer = Whisper = DecodingTask = DecodingOptions = DecodingResult = SuppressTokens = Unavailable
    LANGUAGES = {
        "en": "english",
        "zh": "chinese",
        "de": "german",
        "es": "spanish",
        "ru": "russian",
        "ko": "korean",
        "fr": "french",
        "ja": "japanese",
        "pt": "portuguese",
        "tr": "turkish",
        "pl": "polish",
        "ca": "catalan",
        "nl": "dutch",
        "ar": "arabic",
        "sv": "swedish",
        "it": "italian",
        "id": "indonesian",
        "hi": "hindi",
        "fi": "finnish",
        "vi": "vietnamese",
        "he": "hebrew",
        "uk": "ukrainian",
        "el": "greek",
        "ms": "malay",
        "cs": "czech",
        "ro": "romanian",
        "da": "danish",
        "hu": "hungarian",
        "ta": "tamil",
        "no": "norwegian",
        "th": "thai",
        "ur": "urdu",
        "hr": "croatian",
        "bg": "bulgarian",
        "lt": "lithuanian",
        "la": "latin",
        "mi": "maori",
        "ml": "malayalam",
        "cy": "welsh",
        "sk": "slovak",
        "te": "telugu",
        "fa": "persian",
        "lv": "latvian",
        "bn": "bengali",
        "sr": "serbian",
        "az": "azerbaijani",
        "sl": "slovenian",
        "kn": "kannada",
        "et": "estonian",
        "mk": "macedonian",
        "br": "breton",
        "eu": "basque",
        "is": "icelandic",
        "hy": "armenian",
        "ne": "nepali",
        "mn": "mongolian",
        "bs": "bosnian",
        "kk": "kazakh",
        "sq": "albanian",
        "sw": "swahili",
        "gl": "galician",
        "mr": "marathi",
        "pa": "punjabi",
        "si": "sinhala",
        "km": "khmer",
        "sn": "shona",
        "yo": "yoruba",
        "so": "somali",
        "af": "afrikaans",
        "oc": "occitan",
        "ka": "georgian",
        "be": "belarusian",
        "tg": "tajik",
        "sd": "sindhi",
        "gu": "gujarati",
        "am": "amharic",
        "yi": "yiddish",
        "lo": "lao",
        "uz": "uzbek",
        "fo": "faroese",
        "ht": "haitian creole",
        "ps": "pashto",
        "tk": "turkmen",
        "nn": "nynorsk",
        "mt": "maltese",
        "sa": "sanskrit",
        "lb": "luxembourgish",
        "my": "myanmar",
        "bo": "tibetan",
        "tl": "tagalog",
        "mg": "malagasy",
        "as": "assamese",
        "tt": "tatar",
        "haw": "hawaiian",
        "ln": "lingala",
        "ha": "hausa",
        "ba": "bashkir",
        "jw": "javanese",
        "su": "sundanese",
        "yue": "cantonese",
    }

    # language code lookup by name, with a few language aliases
    TO_LANGUAGE_CODE = {
        **{language: code for code, language in LANGUAGES.items()},
        "burmese": "my",
        "valencian": "ca",
        "flemish": "nl",
        "haitian": "ht",
        "letzeburgesch": "lb",
        "pushto": "ps",
        "panjabi": "pa",
        "moldavian": "ro",
        "moldovan": "ro",
        "sinhalese": "si",
        "castilian": "es",
        "mandarin": "zh",
    }


    def pad_or_trim(array, length: int = N_SAMPLES, *, axis: int = -1):
        """
        Pad or trim the audio array to N_SAMPLES, as expected by the encoder.
        """
        if torch.is_tensor(array):
            if array.shape[axis] > length:
                array = array.index_select(
                    dim=axis, index=torch.arange(length, device=array.device)
                )

            if array.shape[axis] < length:
                pad_widths = [(0, 0)] * array.ndim
                pad_widths[axis] = (0, length - array.shape[axis])
                array = torch.nn.functional.pad(array, [pad for sizes in pad_widths[::-1] for pad in sizes])
        else:
            if array.shape[axis] > length:
                array = array.take(indices=range(length), axis=axis)

            if array.shape[axis] < length:
                pad_widths = [(0, 0)] * array.ndim
                pad_widths[axis] = (0, length - array.shape[axis])
                array = np.pad(array, pad_widths)

        return array


def warn_compatibility_issues(
        whisper_module,
        ignore: bool = False,
        additional_msg: str = ''
):
    compatibility_warning = ''
    if not ignore:
        if whisper_module.__version__ not in _COMPATIBLE_WHISPER_VERSIONS:
            compatibility_warning += (f'Whisper {whisper_module.__version__} is installed.'
                                      f'Versions confirm to be compatible: {", ".join(_COMPATIBLE_WHISPER_VERSIONS)}\n')
        _is_whisper_repo_version = bool(importlib.metadata.distribution('openai-whisper').read_text('direct_url.json'))
        if _is_whisper_repo_version:
            compatibility_warning += ('The detected version appears to be installed from the repository '
                                      'which can have compatibility issues '
                                      'due to multiple commits sharing the same version number. '
                                      f'It is recommended to install version {_required_whisper_ver} from PyPI.\n')

        if compatibility_warning:
            compatibility_warning = (
                    'The installed version of Whisper might be incompatible.\n'
                    + compatibility_warning +
                    'To prevent errors and performance issues, reinstall correct version with: '
                    f'"pip install --upgrade --no-deps --force-reinstall openai-whisper=={_required_whisper_ver}".'
            )
            if additional_msg:
                compatibility_warning += f' {additional_msg}'
            warnings.warn(compatibility_warning)


def get_tokenizer(model=None, is_faster_model: bool = False, **kwargs):
    """
    Backward compatible wrapper of :func:`whisper.tokenizer.get_tokenizer` and
    :class:`faster_whisper.tokenizer.Tokenizer`.
    """
    if is_faster_model:
        import faster_whisper.tokenizer
        tokenizer = faster_whisper.tokenizer.Tokenizer
        params = get_func_parameters(tokenizer)
        if model is not None and 'tokenizer' not in kwargs:
            kwargs['tokenizer'] = model.hf_tokenizer
    else:
        tokenizer = whisper.tokenizer.get_tokenizer
        params = _TOKENIZER_PARAMS
    if model is not None and 'multilingual' not in kwargs:
        kwargs['multilingual'] = \
            (model.is_multilingual if hasattr(model, 'is_multilingual') else model.model.is_multilingual)
    if 'num_languages' in params:
        if hasattr(model, 'num_languages'):
            kwargs['num_languages'] = \
                (model.num_languages if hasattr(model, 'num_languages') else model.model.num_languages)
    elif 'num_languages' in kwargs:
        del kwargs['num_languages']
    return tokenizer(**kwargs)



from typing import Union, Optional

import torch

from .utils import load_audio, SAMPLE_RATE
from ..audio.utils import resample


def is_noisereduce_available():
    from importlib.util import find_spec
    if find_spec('noisereduce') is None:
        raise ModuleNotFoundError("Please install noisereduce; "
                                  "'pip install -U noisereduce'"
                                  "Official repo: https://github.com/timsainb/noisereduce")


def load_noisereduce_model(cache: bool = True, **kwargs):
    is_noisereduce_available()
    from noisereduce import reduce_noise

    class NRWrapper:
        samplerate = SAMPLE_RATE

        def __call__(self, audio, **kwargs):
            options = dict(
                n_fft=512,
                stationary=False,
                use_torch=True,
                sr=self.samplerate
            )
            options.update(kwargs)
            return torch.from_numpy(reduce_noise(y=audio, **options))

    model = NRWrapper()
    return model


def noisereduce_audio(
    audio: Union[torch.Tensor, str, bytes],
    input_sr: int = None,
    output_sr: int = None,
    model=None,
    device: str = "cuda",
    verbose: bool = True,
    save_path: Optional[Union[str, callable]] = None,
    **noisereduce_options
) -> torch.Tensor:
    """
    Remove noise from ``audio`` with noisereduce.

    For list of options, see repo: https://github.com/timsainb/noisereduce.
    """
    if model is None:
        model = load_noisereduce_model()
    if noisereduce_options.get("sr", input_sr) != input_sr:
        raise ValueError('Specified conflicting ``input_sr`` and ``sr``.')
    if isinstance(audio, (str, bytes)):
        audio = torch.from_numpy(load_audio(audio, model.samplerate))
        input_sr = model.samplerate
    elif input_sr != model.samplerate:
        if input_sr is None:
            raise ValueError('No ``input_sr`` specified for audio tensor.')
        audio = resample(audio, input_sr, model.samplerate)
    audio_dims = audio.dim()
    assert audio_dims <= 2
    if dims_missing := 2 - audio_dims:
        audio = audio[[None]*dims_missing]
    if audio.shape[-2] == 1:
        audio = audio.repeat_interleave(2, -2)

    noisereduce_options.pop('progress', None)  # not implemented
    denoised_audio = model(audio=audio, device=device, **noisereduce_options).mean(dim=0)

    if 'cuda' in str(device):
        torch.cuda.empty_cache()

    if output_sr is not None and input_sr != output_sr:
        denoised_audio = resample(denoised_audio, input_sr, output_sr)

    if save_path is not None:
        if isinstance(save_path, str):
            from .output import save_audio_tensor
            save_audio_tensor(denoised_audio, save_path, output_sr or input_sr, verbose=verbose)
        else:
            save_path(denoised_audio, output_sr)

    return denoised_audio



from typing import TYPE_CHECKING, List, Union
from dataclasses import replace

import torch
import numpy as np

from .whisper_compatibility import DecodingTask, DecodingOptions, DecodingResult


if TYPE_CHECKING:
    from .whisper_compatibility import Whisper


def _suppress_ts(ts_logits: torch.Tensor, ts_token_mask: torch.Tensor = None):
    if ts_token_mask is not None:
        ts_logits[:, ts_token_mask] = -np.inf


# modified version of whisper.decoding.DecodingTask
class DecodingTaskStable(DecodingTask):

    def __init__(self, *args, **kwargs):
        self.ts_token_mask: torch.Tensor = kwargs.pop('ts_token_mask', None)
        self.audio_features: torch.Tensor = kwargs.pop('audio_features', None)
        super(DecodingTaskStable, self).__init__(*args, **kwargs)

    def _get_audio_features(self, mel: torch.Tensor):
        if self.audio_features is None:
            audio_features = super()._get_audio_features(mel)
            self.audio_features = audio_features.detach().clone()
            return audio_features
        return self.audio_features.clone()

    # modified version of whisper.DecodingTask._main_loop
    def _main_loop(self, audio_features: torch.Tensor, tokens: torch.Tensor):
        n_batch = tokens.shape[0]
        sum_logprobs: torch.Tensor = torch.zeros(n_batch, device=audio_features.device)
        no_speech_probs = [np.nan] * n_batch

        try:
            for i in range(self.sample_len):
                logits = self.inference.logits(tokens, audio_features)

                if i == 0 and self.tokenizer.no_speech is not None:  # save no_speech_probs
                    probs_at_sot = logits[:, self.sot_index].float().softmax(dim=-1)
                    no_speech_probs = probs_at_sot[:, self.tokenizer.no_speech].tolist()

                # now we need to consider the logits at the last token only
                logits = logits[:, -1]

                # apply the logit filters, e.g. for suppressing or applying penalty to
                for logit_filter in self.logit_filters:
                    logit_filter.apply(logits, tokens)

                # suppress timestamp tokens where the audio is silent so that decoder ignores those timestamps
                _suppress_ts(logits[:, self.tokenizer.timestamp_begin:], self.ts_token_mask)

                logits.nan_to_num_(-np.inf)
                # expand the tokens tensor with the selected next tokens
                tokens, completed = self.decoder.update(tokens, logits, sum_logprobs)

                if completed or tokens.shape[-1] > self.n_ctx:
                    break
        finally:
            self.inference.cleanup_caching()

        return tokens, sum_logprobs, no_speech_probs


# modified version of whisper.decoding.decode
@torch.no_grad()
def decode_stable(model: "Whisper",
                  mel: torch.Tensor,
                  options: DecodingOptions = None,
                  ts_token_mask: torch.Tensor = None,
                  audio_features: torch.Tensor = None,
                  **kwargs, ) -> \
        Union[DecodingResult, List[DecodingResult], tuple]:
    """
    Performs decoding of 30-second audio segment(s), provided as Mel spectrogram(s).

    Parameters
    ----------
    model : whisper.model.Whisper
        An instance of Whisper ASR model.
    mel : torch.Tensor,
        A tensor containing the Mel spectrogram(s). ``mel.shape`` must be (80, 3000) or (*, 80, 3000).
    options : whisper.decode.DecodingOptions, default whisper.decode.DecodingOptions()
        A dataclass that contains all necessary options for decoding 30-second segments
    ts_token_mask : torch.Tensor, optional
        Mask for suppressing to timestamp token(s) for decoding.
    audio_features : torch.Tensor, optional
        Reused ``audio_feature`` from encoder for fallback.

    Returns
    -------
    whisper.decode.DecodingResult or list whisper.decode.DecodingResult
        The result(s) of decoding contained in ``whisper.decode.DecodingResult`` dataclass instance(s).
    """
    if single := mel.ndim == 2:
        mel = mel.unsqueeze(0)

    if options is None:
        options = DecodingOptions()

    if kwargs:
        options = replace(options, **kwargs)

    task = DecodingTaskStable(model, options, ts_token_mask=ts_token_mask, audio_features=audio_features)
    result = task.run(mel)

    return result[0] if single else result, task.audio_features



from types import MethodType
from typing import Union, Optional, Callable

import numpy as np
from tqdm import tqdm

from ..result import Segment, WhisperResult
from ..non_whisper import transcribe_any
from ..utils import safe_print, isolate_useful_options
from ..audio import audioloader_not_supported, convert_demucs_kwargs

from ..whisper_compatibility import LANGUAGES


def faster_transcribe(
        model: "WhisperModel",
        audio: Union[str, bytes, np.ndarray],
        *,
        word_timestamps: bool = True,
        verbose: Optional[bool] = False,
        regroup: Union[bool, str] = True,
        suppress_silence: bool = True,
        suppress_word_ts: bool = True,
        use_word_position: bool = True,
        q_levels: int = 20,
        k_size: int = 5,
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: bool = False,
        demucs_options: dict = None,
        vad: bool = False,
        vad_threshold: float = 0.35,
        vad_onnx: bool = False,
        min_word_dur: Optional[float] = None,
        min_silence_dur: Optional[float] = None,
        nonspeech_error: float = 0.1,
        only_voice_freq: bool = False,
        only_ffmpeg: bool = False,
        check_sorted: bool = True,
        progress_callback: Callable = None,
        **options
) -> WhisperResult:
    """
    Transcribe audio using faster-whisper (https://github.com/guillaumekln/faster-whisper).

    This is uses the transcribe method from faster-whisper, :meth:`faster_whisper.WhisperModel.transcribe`, while
    still allowing additional preprocessing and postprocessing. The preprocessing performed on the audio includes:
    voice isolation / noise removal and low/high-pass filter. The postprocessing performed on the
    transcription result includes: adjusting timestamps with VAD and custom regrouping segments based punctuation
    and speech gaps.

    Parameters
    ----------
    model : faster_whisper.WhisperModel
        The faster-whisper ASR model instance.
    audio : str or numpy.ndarray or torch.Tensor or bytes
        Path/URL to the audio file, the audio waveform, or bytes of audio file.
        If audio is :class:`numpy.ndarray` or :class:`torch.Tensor`, the audio must be already at sampled to 16kHz.
    verbose : bool or None, default False
        Whether to display the text being decoded to the console.
        Displays all the details if ``True``. Displays progressbar if ``False``. Display nothing if ``None``.
    word_timestamps : bool, default True
        Extract word-level timestamps using the cross-attention pattern and dynamic time warping,
        and include the timestamps for each word in each segment.
        Disabling this will prevent segments from splitting/merging properly.
    regroup : bool or str, default True, meaning the default regroup algorithm
        String for customizing the regrouping algorithm. False disables regrouping.
        Ignored if ``word_timestamps = False``.
    suppress_silence : bool, default True
        Whether to enable timestamps adjustments based on the detected silence.
    suppress_word_ts : bool, default True
        Whether to adjust word timestamps based on the detected silence. Only enabled if ``suppress_silence = True``.
    use_word_position : bool, default True
        Whether to use position of the word in its segment to determine whether to keep end or start timestamps if
        adjustments are required. If it is the first word, keep end. Else if it is the last word, keep the start.
    q_levels : int, default 20
        Quantization levels for generating timestamp suppression mask; ignored if ``vad = true``.
        Acts as a threshold to marking sound as silent.
        Fewer levels will increase the threshold of volume at which to mark a sound as silent.
    k_size : int, default 5
        Kernel size for avg-pooling waveform to generate timestamp suppression mask; ignored if ``vad = true``.
        Recommend 5 or 3; higher sizes will reduce detection of silence.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    vad : bool, default False
        Whether to use Silero VAD to generate timestamp suppression mask.
        Silero VAD requires PyTorch 1.12.0+. Official repo, https://github.com/snakers4/silero-vad.
    vad_threshold : float, default 0.35
        Threshold for detecting speech with Silero VAD. Low threshold reduces false positives for silence detection.
    vad_onnx : bool, default False
        Whether to use ONNX for Silero VAD.
    min_word_dur : float or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Shortest duration each word is allowed to reach for silence suppression.
    min_silence_dur : float, optional
        Shortest duration of silence allowed for silence suppression.
    nonspeech_error : float, default 0.3
        Relative error of non-speech sections that appear in between a word for silence suppression.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    only_ffmpeg : bool, default False
        Whether to use only FFmpeg (instead of not yt-dlp) for URls
    check_sorted : bool, default True
        Whether to raise an error when timestamps returned by faster-whipser are not in ascending order.
    progress_callback : Callable, optional
        A function that will be called when transcription progress is updated.
        The callback need two parameters.
        The first parameter is a float for seconds of the audio that has been transcribed.
        The second parameter is a float for total duration of audio in seconds.
    options
        Additional options used for :meth:`faster_whisper.WhisperModel.transcribe` and
        :func:`stable_whisper.non_whisper.transcribe_any`.

    Returns
    -------
    stable_whisper.result.WhisperResult
        All timestamps, words, probabilities, and other data from the transcription of ``audio``.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_faster_whisper('base')
    >>> result = model.transcribe_stable('audio.mp3', vad=True)
    >>> result.to_srt_vtt('audio.srt')
    Saved: audio.srt
    """
    audioloader_not_supported(audio)
    extra_options = isolate_useful_options(options, transcribe_any, pop=True)
    denoiser, denoiser_options = convert_demucs_kwargs(
        denoiser, denoiser_options, demucs=demucs, demucs_options=demucs_options
    )
    if not isinstance(audio, (str, bytes)):
        if 'input_sr' not in extra_options:
            extra_options['input_sr'] = model.feature_extractor.sampling_rate

    if denoiser or only_voice_freq:
        if 'audio_type' not in extra_options:
            extra_options['audio_type'] = 'numpy'
        if 'model_sr' not in extra_options:
            extra_options['model_sr'] = model.feature_extractor.sampling_rate
    faster_whisper_options = options
    faster_whisper_options['model'] = model
    faster_whisper_options['audio'] = audio
    faster_whisper_options['word_timestamps'] = word_timestamps
    faster_whisper_options['verbose'] = verbose
    faster_whisper_options['progress_callback'] = progress_callback

    return transcribe_any(
        inference_func=_inner_transcribe,
        audio=audio,
        inference_kwargs=faster_whisper_options,
        verbose=verbose,
        regroup=regroup,
        suppress_silence=suppress_silence,
        suppress_word_ts=suppress_word_ts,
        q_levels=q_levels,
        k_size=k_size,
        denoiser=denoiser,
        denoiser_options=denoiser_options,
        vad=vad,
        vad_threshold=vad_threshold,
        vad_onnx=vad_onnx,
        min_word_dur=min_word_dur,
        min_silence_dur=min_silence_dur,
        nonspeech_error=nonspeech_error,
        use_word_position=use_word_position,
        only_voice_freq=only_voice_freq,
        only_ffmpeg=only_ffmpeg,
        force_order=True,
        check_sorted=check_sorted,
        **extra_options
    )


def _inner_transcribe(model, audio, verbose, **faster_transcribe_options):
    if isinstance(audio, bytes):
        import io
        audio = io.BytesIO(audio)
    progress_callback = faster_transcribe_options.pop('progress_callback', None)
    segments, info = model.transcribe(audio, **faster_transcribe_options)
    language = LANGUAGES.get(info.language, info.language)
    if verbose is not None:
        print(f'Detected Language: {language}')
        print(f'Transcribing with faster-whisper ({model.model_size_or_path})...\r', end='')

    final_segments = []
    task = faster_transcribe_options.get('task', 'transcribe').title()
    total_duration = round(info.duration, 2)

    with tqdm(total=total_duration, unit='sec', disable=verbose is not False, desc=task) as tqdm_pbar:

        def update_pbar(seek):
            tqdm_pbar.update(seek - tqdm_pbar.n)
            if progress_callback is not None:
                progress_callback(seek, total_duration)

        for segment in segments:
            segment = segment._asdict()
            if (words := segment.get('words')) is not None:
                segment['words'] = [w._asdict() for w in words]
            else:
                del segment['words']
            if verbose:
                safe_print(Segment(**segment, ignore_unused_args=True).to_display_str())
            final_segments.append(segment)
            update_pbar(segment["end"])
        update_pbar(tqdm_pbar.total)

    if verbose:
        print(f'Completed transcription with faster-whisper ({model.model_size_or_path}).')

    return dict(language=language, segments=final_segments)


def load_faster_whisper(model_size_or_path: str, **model_init_options):
    """
    Load an instance of :class:`faster_whisper.WhisperModel`.

    Parameters
    ----------
    model_size_or_path : {'tiny', 'tiny.en', 'base', 'base.en', 'small', 'small.en', 'medium', 'medium.en', 'large-v1',
        'large-v2', 'large-v3', or 'large'}
        Size of the model.

    model_init_options
        Additional options to use for initialization of :class:`faster_whisper.WhisperModel`.

    Returns
    -------
    faster_whisper.WhisperModel
        A modified instance with :func:`stable_whisper.whisper_word_level.faster_whisper.faster_transcribe`
        assigned to :meth:`faster_whisper.WhisperModel.transcribe_stable`.
    """
    from faster_whisper import WhisperModel
    faster_model = WhisperModel(model_size_or_path, **model_init_options)
    faster_model.model_size_or_path = model_size_or_path

    faster_model.transcribe_stable = MethodType(faster_transcribe, faster_model)
    from ..alignment import align
    faster_model.align = MethodType(align, faster_model)

    return faster_model



import copy
import re
import warnings

import torch
import numpy as np
from tqdm import tqdm
from typing import TYPE_CHECKING, Union, List, Callable, Optional, Tuple

from .result import WhisperResult, Segment
from .timing import add_word_timestamps_stable, split_word_tokens
from .audio import prep_audio, AudioLoader, audioloader_not_supported, convert_demucs_kwargs
from .utils import safe_print, format_timestamp
from .whisper_compatibility import warn_compatibility_issues, get_tokenizer
from .stabilization import NonSpeechPredictor
from .default import get_min_word_dur, get_prepend_punctuations, get_append_punctuations

from .whisper_compatibility import (
    SAMPLE_RATE, N_FRAMES, N_FFT, pad_or_trim, log_mel_spectrogram, FRAMES_PER_SECOND, CHUNK_LENGTH, N_SAMPLES,
    median_filter, DecodingTask, DecodingOptions, SuppressTokens, whisper, TOKENS_PER_SECOND
)

if TYPE_CHECKING:
    from .whisper_compatibility import Whisper
    from .whisper_compatibility import Tokenizer

__all__ = ['align', 'refine', 'locate']


def align(
        model: "Whisper",
        audio: Union[str, np.ndarray, torch.Tensor, bytes, AudioLoader],
        text: Union[str, List[int], WhisperResult],
        language: str = None,
        *,
        verbose: Optional[bool] = False,
        regroup: bool = True,
        suppress_silence: bool = True,
        suppress_word_ts: bool = True,
        suppress_attention: bool = False,
        use_word_position: bool = True,
        min_word_dur: Optional[float] = None,
        min_silence_dur: Optional[float] = None,
        nonspeech_error: float = 0.1,
        q_levels: int = 20,
        k_size: int = 5,
        vad: bool = False,
        vad_threshold: float = 0.35,
        vad_onnx: bool = False,
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: Union[bool, torch.nn.Module] = False,
        demucs_options: dict = None,
        only_voice_freq: bool = False,
        prepend_punctuations: Optional[str] = None,
        append_punctuations: Optional[str] = None,
        progress_callback: Callable = None,
        ignore_compatibility: bool = False,
        remove_instant_words: bool = False,
        token_step: int = 100,
        original_split: bool = False,
        word_dur_factor: Optional[float] = 2.0,
        max_word_dur: Optional[float] = 3.0,
        nonspeech_skip: Optional[float] = 5.0,
        fast_mode: bool = False,
        tokenizer: "Tokenizer" = None,
        stream: Optional[bool] = None,
        failure_threshold: Optional[float] = None,
        extra_models: Optional[List["Whisper"]] = None,
        presplit: Union[bool, List[str]] = True,
        gap_padding: str = ' ...'
) -> Union[WhisperResult, None]:
    """
    Align plain text or tokens with audio at word-level.

    Since this is significantly faster than transcribing, it is a more efficient method for testing various settings
    without re-transcribing. This is also useful for timing a more correct transcript than one that Whisper can produce.

    Parameters
    ----------
    model : "Whisper"
        The Whisper ASR model modified instance
    audio : str or numpy.ndarray or torch.Tensor or bytes or AudioLoader
        Path/URL to the audio file, the audio waveform, or bytes of audio file or
        instance of :class:`stable_whisper.audio.AudioLoader`.
        If audio is :class:`numpy.ndarray` or :class:`torch.Tensor`, the audio must be already at sampled to 16kHz.
    text : str or list of int or stable_whisper.result.WhisperResult
        String of plain-text, list of tokens, or instance of :class:`stable_whisper.result.WhisperResult`.
    language : str, default None, uses ``language`` in ``text`` if it is a :class:`stable_whisper.result.WhisperResult`
        Language of ``text``. Required if ``text`` does not contain ``language``.
    remove_instant_words : bool, default False
        Whether to truncate any words with zero duration.
    token_step : int, default 100
        Max number of tokens to align each pass. Use higher values to reduce chance of misalignment.
    original_split : bool, default False
        Whether to preserve the original segment groupings. Segments are spit by line break if ``text`` is plain-text.
    max_word_dur : float or None, default 3.0
        Global maximum word duration in seconds. Re-align words that exceed the global maximum word duration.
    word_dur_factor : float or None, default 2.0
        Factor to compute the Local maximum word duration, which is ``word_dur_factor`` * local medium word duration.
        Words that need re-alignment, are re-algined with duration <= local/global maximum word duration.
    nonspeech_skip : float or None, default 5.0
        Skip non-speech sections that are equal or longer than this duration in seconds. Disable skipping if ``None``.
    fast_mode : bool, default False
        Whether to speed up alignment by re-alignment with local/global maximum word duration.
        ``True`` tends produce better timestamps when ``text`` is accurate and there are no large speechless gaps.
    tokenizer : "Tokenizer", default None, meaning a new tokenizer is created according ``language`` and ``model``
        A tokenizer to used tokenizer text and detokenize tokens.
    stream : bool or None, default None
        Whether to loading ``audio`` in chunks of 30 seconds until the end of file/stream.
        If ``None`` and ``audio`` is a string then set to ``True`` else ``False``.
    failure_threshold : float, optional
        Abort alignment when percentage of words with zero duration exceeds ``failure_threshold``.
    verbose : bool or None, default False
        Whether to display the text being decoded to the console.
        Displays all the details if ``True``. Displays progressbar if ``False``. Display nothing if ``None``.
    regroup : bool or str, default True, meaning the default regroup algorithm
        String for customizing the regrouping algorithm. False disables regrouping.
        Ignored if ``word_timestamps = False``.
    suppress_silence : bool, default True
        Whether to enable timestamps adjustments based on the detected silence.
    suppress_word_ts : bool, default True
        Whether to adjust word timestamps based on the detected silence. Only enabled if ``suppress_silence = True``.
    use_word_position : bool, default True
        Whether to use position of the word in its segment to determine whether to keep end or start timestamps if
        adjustments are required. If it is the first word, keep end. Else if it is the last word, keep the start.
    q_levels : int, default 20
        Quantization levels for generating timestamp suppression mask; ignored if ``vad = true``.
        Acts as a threshold to marking sound as silent.
        Fewer levels will increase the threshold of volume at which to mark a sound as silent.
    k_size : int, default 5
        Kernel size for avg-pooling waveform to generate timestamp suppression mask; ignored if ``vad = true``.
        Recommend 5 or 3; higher sizes will reduce detection of silence.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    vad : bool, default False
        Whether to use Silero VAD to generate timestamp suppression mask.
        Silero VAD requires PyTorch 1.12.0+. Official repo, https://github.com/snakers4/silero-vad.
    vad_threshold : float, default 0.35
        Threshold for detecting speech with Silero VAD. Low threshold reduces false positives for silence detection.
    vad_onnx : bool, default False
        Whether to use ONNX for Silero VAD.
    min_word_dur : float or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Shortest duration each word is allowed to reach for silence suppression.
    min_silence_dur : float, optional
        Shortest duration of silence allowed for silence suppression.
    nonspeech_error : float, default 0.1
        Relative error of non-speech sections that appear in between a word for silence suppression.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    prepend_punctuations : str or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Punctuations to prepend to next word.
    append_punctuations : str or None, default None meaning use ``stable_whisper.default.DEFAULT_VALUES``
        Punctuations to append to previous word.
    progress_callback : Callable, optional
        A function that will be called when transcription progress is updated.
        The callback need two parameters.
        The first parameter is a float for seconds of the audio that has been transcribed.
        The second parameter is a float for total duration of audio in seconds.
    ignore_compatibility : bool, default False
        Whether to ignore warnings for compatibility issues with the detected Whisper version.
    extra_models : list of whisper.model.Whisper, optional
        List of additional Whisper model instances to use for computing word-timestamps along with ``model``.
    presplit : bool or list of str, default True meaning ['.', '。', '?', '？']
        List of ending punctuation used to split ``text`` into segments for applying ``gap_padding``,
        but segmentation of final output is unnaffected unless ``original_split=True``.
        If ``original_split=True``, the original split is used instead of split from ``presplit``.
        Ignored if ``model`` is a faster-whisper model.
    gap_padding : str, default ' ...'
        Only if ``presplit=True``, ``gap_padding`` is prepended to each segments for word timing alignment.
        Used to reduce the probability of model predicting timestamps earlier than the first utterance.
        Ignored if ``model`` is a faster-whisper model.

    Returns
    -------
    stable_whisper.result.WhisperResult or None
        All timestamps, words, probabilities, and other data from the alignment of ``audio``. Return None if alignment
        fails and ``remove_instant_words = True``.

    Notes
    -----
    If ``token_step`` is less than 1, ``token_step`` will be set to its maximum value, 442. This value is computed with
    ``whisper.model.Whisper.dims.n_text_ctx`` - 6.

    IF ``original_split = True`` and a line break is found in middle of a word in ``text``, the split will occur after
    that word.

    ``regroup`` is ignored if ``original_split = True``.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.align('helloworld.mp3', 'Hello, World!', 'English')
    >>> result.to_srt_vtt('helloword.srt')
    Saved 'helloworld.srt'
    """
    if suppress_attention:
        warnings.warn('``suppress_attention`` is deprecated and will be removed in future versions',
                      stacklevel=2)
    denoiser, denoiser_options = convert_demucs_kwargs(
        denoiser, denoiser_options, demucs=demucs, demucs_options=demucs_options
    )
    prepend_punctuations = get_prepend_punctuations(prepend_punctuations)
    append_punctuations = get_append_punctuations(append_punctuations)
    min_word_dur = get_min_word_dur(min_word_dur)
    if failure_threshold is not None and (failure_threshold < 0 or failure_threshold > 1):
        raise ValueError(f'``failure_threshold`` ({failure_threshold}) must be between 0 and 1.')
    is_faster_model = model.__module__.startswith('faster_whisper.')
    if not is_faster_model:
        warn_compatibility_issues(whisper, ignore_compatibility)
    max_token_step = (model.max_length if is_faster_model else model.dims.n_text_ctx) - 6
    if token_step < 1:
        token_step = max_token_step
    elif token_step > max_token_step:
        raise ValueError(f'The max value for [token_step] is {max_token_step} but got {token_step}.')

    split_indices_by_char = []
    if isinstance(text, WhisperResult):
        if language is None:
            language = text.language
        if original_split and len(text.segments) > 1 and text.has_words:
            split_indices_by_char = np.cumsum([sum(len(w.word) for w in seg.words) for seg in text.segments])
        text = text.all_tokens() if text.has_words and all(w.tokens for w in text.all_words()) else text.text
    elif isinstance(text, str):
        if original_split and '\n' in text:
            text_split = [
                ' '+norm_line
                for line in text.splitlines()
                if (norm_line := re.sub(r'\s', ' ', line).strip())
            ]
            split_indices_by_char = np.cumsum([len(seg) for seg in text_split])
            text = ''.join(seg for seg in text_split)
        else:
            text = re.sub(r'\s', ' ', text)
            if not text.startswith(' '):
                text = ' ' + text
    if language is None:
        raise TypeError('expected argument for language')
    if tokenizer is None:
        tokenizer = get_tokenizer(model, is_faster_model=is_faster_model, language=language, task='transcribe')
    tokens = tokenizer.encode(text) if isinstance(text, str) else text
    tokens = [t for t in tokens if t < tokenizer.eot]
    _, (words, word_tokens), _ = split_word_tokens([dict(tokens=tokens)], tokenizer)
    pad_mask = None
    if is_faster_model:
        presplit = False
    if presplit:
        if not isinstance(presplit, List):
            presplit = ['.', '。', '?', '？']
        if len(split_indices_by_char):
            pad_mask = []
            cumsums = split_indices_by_char.tolist()
            cumsum_len = 0
            for word in words:
                cumsum_len += len(word)
                if cumsums and cumsum_len >= cumsums[0]:
                    cumsums.pop(0)
                    pad_mask.extend([True]*len(word))
                else:
                    pad_mask.extend([False]*len(word))
        else:
            pad_mask = [b for w in words for b in [any(map(w.endswith, presplit))]*len(w)]

    if isinstance(audio, AudioLoader):
        audio.validate_external_args(
            sr=SAMPLE_RATE,
            vad=vad,
            stream=stream,
            denoiser=denoiser,
            denoiser_options=denoiser_options,
            only_voice_freq=only_voice_freq
        )
    else:
        audio = AudioLoader(
            audio,
            sr=SAMPLE_RATE,
            denoiser=denoiser,
            denoiser_options=denoiser_options,
            only_voice_freq=only_voice_freq,
            verbose=verbose,
            new_chunk_divisor=512,
            stream=stream
        )

    initial_duration = audio.get_duration(2)

    seek_sample = 0
    total_words = len(words)
    remaining_len = sum(len(w) for w in words)

    if is_faster_model:

        def timestamp_words():
            temp_segment = dict(
                seek=0,
                start=0.0,
                end=round(segment_samples / model.feature_extractor.sampling_rate, 3),
                tokens=[t for wt in curr_word_tokens for t in wt],
            )
            features = model.feature_extractor(audio_segment.cpu().numpy())
            encoder_output = model.encode(features[:, : model.feature_extractor.nb_max_frames])

            model.add_word_timestamps(
                segments=[temp_segment],
                tokenizer=tokenizer,
                encoder_output=encoder_output,
                num_frames=round(segment_samples / model.feature_extractor.hop_length),
                prepend_punctuations=prepend_punctuations,
                append_punctuations=append_punctuations,
                last_speech_timestamp=temp_segment['start'],
            )

            cumsum_lens = np.cumsum([len(w) for w in curr_words]).tolist()
            final_cumsum_lens = np.cumsum([len(w['word']) for w in temp_segment['words']]).tolist()

            assert not (set(final_cumsum_lens) - set(cumsum_lens)), 'word mismatch'
            prev_l_idx = 0
            for w_idx, cs_len in enumerate(final_cumsum_lens):
                temp_segment['words'][w_idx]['start'] = round(temp_segment['words'][w_idx]['start'] + time_offset, 3)
                temp_segment['words'][w_idx]['end'] = round(temp_segment['words'][w_idx]['end'] + time_offset, 3)
                l_idx = cumsum_lens.index(cs_len)+1
                temp_segment['words'][w_idx]['tokens'] = [t for wt in curr_word_tokens[prev_l_idx:l_idx] for t in wt]
                prev_l_idx = l_idx

            return temp_segment

    else:
        def timestamp_words():
            if curr_split_indices:
                temp_split_indices = [0] + curr_split_indices
                if temp_split_indices[-1] < len(curr_words):
                    temp_split_indices.append(len(curr_words))
                temp_segments = [
                    dict(
                        seek=time_offset,
                        tokens=(curr_words[i:j], curr_word_tokens[i:j])
                    )
                    for i, j in zip(temp_split_indices[:-1], temp_split_indices[1:])
                ]
            else:
                temp_segments = [dict(seek=time_offset, tokens=(curr_words, curr_word_tokens))]
            sample_padding = max(N_SAMPLES - audio_segment.shape[-1], 0)
            mel_segment = log_mel_spectrogram(audio_segment, model.dims.n_mels, padding=sample_padding)
            mel_segment = pad_or_trim(mel_segment, N_FRAMES).to(device=model.device)

            add_word_timestamps_stable(
                segments=temp_segments,
                model=model,
                tokenizer=tokenizer,
                mel=mel_segment,
                num_samples=segment_samples,
                split_callback=(lambda x, _: x),
                prepend_punctuations=prepend_punctuations,
                append_punctuations=append_punctuations,
                gap_padding=gap_padding if presplit else None,
                extra_models=extra_models,
                pad_first_seg=pad_first_seg
            )
            if len(temp_segments) == 1:
                return temp_segments[0]
            return dict(words=[w for seg in temp_segments for w in seg['words']])

    def get_curr_words():
        nonlocal words, word_tokens, pad_mask, remaining_len
        curr_tk_count = 0
        w, wt, m = [], [], []
        is_start_gap = (True if remaining_len == len(pad_mask) else pad_mask[-(remaining_len+1)]) if pad_mask else True
        for i in range(len(words)):
            tk_count = len(word_tokens[0])
            m_count = 1 if pad_mask and pad_mask[-(remaining_len-len(words[0])+1)] else 0
            if curr_tk_count + len(m) + tk_count + m_count > token_step and w:
                break
            if pad_mask and pad_mask[-(remaining_len-len(words[0])+1)]:
                    m.append(i+1)
            remaining_len -= len(words[0])
            w.append(words.pop(0))
            wt.append(word_tokens.pop(0))
            curr_tk_count += tk_count
        return w, wt, m, is_start_gap
    result = []

    nonspeech_predictor = NonSpeechPredictor(
        vad=vad if suppress_silence else None,
        mask_pad_func=pad_or_trim,
        get_mask=True,
        min_word_dur=min_word_dur,
        q_levels=q_levels,
        k_size=k_size,
        vad_threshold=vad_threshold,
        vad_onnx=vad_onnx,
        vad_window=audio.new_chunk_divisor,
        sampling_rate=SAMPLE_RATE,
        verbose=None if audio.stream else verbose,
        store_timings=True,
        ignore_is_silent=True,
        min_silence_dur=min_silence_dur
    )
    audio.update_post_prep_callback(nonspeech_predictor.get_on_prep_callback(audio.stream))
    failure_count, max_fail = 0, total_words * (failure_threshold or 1)
    all_punctuations = prepend_punctuations + append_punctuations

    def fix_temp_words(target_word: dict, word_sources: List[dict], second_target: Optional[dict] = None):
        first_word_src = word_sources[0]
        assert target_word['word'].startswith(first_word_src['word'])
        if target_word['word'] != first_word_src['word']:
            if len(word_sources) < 2:
                return None, []
            first_word_src['probability'] = [first_word_src['probability']]
            if first_word_src['word'].strip() in all_punctuations:
                first_word_src['start'], first_word_src['end'] = word_sources[1]['start'], word_sources[1]['end']
            for _ in range(len(word_sources)-1):
                tw = word_sources.pop(1)
                fullword = first_word_src['word'] + tw['word']
                assert target_word['word'].startswith(fullword)
                first_word_src['word'] = fullword
                first_word_src['tokens'] += tw['tokens']
                first_word_src['probability'].append(tw['probability'])
                if tw['word'].strip() not in all_punctuations:
                    first_word_src['end'] = tw['end']
                if target_word['word'] == first_word_src['word']:
                    break
            if target_word['word'] != first_word_src['word']:
                return None, []
            if isinstance(first_word_src['probability'], list):
                first_word_src['probability'] = np.mean(first_word_src['probability']).item()
        elif second_target:
            if len(word_sources) == 1:
                return first_word_src, []
            second_word_src, word_sources = fix_temp_words(second_target, word_sources[1:])
            if second_word_src is not None:
                word_sources = [second_word_src] + word_sources
            return first_word_src, word_sources

        return first_word_src, word_sources[1:]

    def speech_percentage(_word: dict, _mask: torch.Tensor, _offset: float):
        if _mask is None:
            return 1
        s, e = _word['start'], _word['end']
        s = int((s - _offset) * TOKENS_PER_SECOND)
        e = int((e - _offset) * TOKENS_PER_SECOND)
        return 1 - _mask[s:e].float().mean().nan_to_num()

    def is_new_better(w0, m0, o0, w1, m1, o1):
        speech0 = speech_percentage(w0, m0, o0).round(decimals=1)
        speech1 = speech_percentage(w1, m1, o1).round(decimals=1)
        w0p, w1p = w0['probability'], w1['probability']
        return ((w1p**0.75 - w0p**0.75) < 0.35 and speech0 >= speech1) or w0p >= w1p

    with tqdm(total=initial_duration, unit='sec', disable=verbose is not False, desc='Align') as tqdm_pbar:

        def update_pbar(finish: bool = False):
            curr_total = audio.get_duration(2)
            if need_refresh := curr_total != tqdm_pbar.total:
                tqdm_pbar.total = curr_total
            tqdm_pbar.update((curr_total if finish else min(round(last_ts, 2), curr_total)) - tqdm_pbar.n)
            if need_refresh:
                tqdm_pbar.refresh()
            if progress_callback is not None:
                progress_callback(seek=tqdm_pbar.n, total=tqdm_pbar.total)

        def update_curr_words():
            if temp_data['temp_word'] is not None:
                temp_words = [temp_data['temp_word']] + temp_data['extra_words'][:len(curr_words) - 1]
                curr_words[:len(temp_words)] = temp_words
                temp_data['temp_word'] = None

        def redo_words(_idx: int = None):
            nonlocal seg_words, seg_tokens, seg_words, words, word_tokens, curr_words, temp_data, remaining_len
            if _idx is not None and curr_words and temp_data['temp_word'] is not None:
                temp_data['temp_word'], temp_data['extra_words'] = fix_temp_words(
                    curr_words[0],
                    [temp_data['temp_word']] + temp_data['extra_words'],
                    curr_words[1] if len(curr_words) > 1 else None
                )

                if temp_data['temp_word']:
                    use_new = is_new_better(
                        curr_words[0], nonspeech_preds['mask'], time_offset,
                        temp_data['temp_word'], temp_data['temp_mask'], temp_data['temp_offset']
                    )
                    new_extra_words = []
                    if use_new:
                        temp_data['temp_word'] = curr_words[0]
                    else:
                        for wi, (cw, tw) in enumerate(zip(curr_words[1:], temp_data['extra_words'])):
                            assert cw['word'].startswith(tw['word'])
                            use_new = is_new_better(
                                cw, nonspeech_preds['mask'], time_offset,
                                tw, temp_data['temp_mask'], temp_data['temp_offset']
                            )
                            if use_new or cw['word'] != tw['word'] or cw['end'] < tw['end']:
                                break
                            new_extra_words.append(tw)
                    temp_data['extra_words'] = new_extra_words

            if _idx is None:  # redo all
                remaining_len += sum(len(w) for w in seg_words)
                words = seg_words + words
                word_tokens = seg_tokens + word_tokens
                curr_words = []
            elif _idx != len(seg_words):  # redo from _idx
                remaining_len += sum(len(w) for w in seg_words[_idx:])
                words = seg_words[_idx:] + words
                word_tokens = seg_tokens[_idx:] + word_tokens
                curr_words, new_extra_words = curr_words[:_idx], curr_words[_idx:]
                if curr_words:
                    update_curr_words()
                    remaining_len += sum(len(w) for w in seg_words[_idx-1:_idx])
                    words = seg_words[_idx-1:_idx] + words
                    word_tokens = seg_tokens[_idx-1:_idx] + word_tokens
                    temp_data['temp_word'] = curr_words.pop(-1)
                    temp_data['extra_words'] = new_extra_words
                    temp_data['temp_mask'] = nonspeech_preds['mask']
                    temp_data['temp_offset'] = time_offset
            else:
                update_curr_words()

        temp_data: dict = dict(temp_word=None)

        while words:

            time_offset = seek_sample / SAMPLE_RATE
            audio_segment = audio.next_chunk(seek_sample, N_SAMPLES)
            if audio_segment is None:
                break

            segment_samples = audio_segment.shape[-1]
            curr_total_samples = audio.get_total_samples()

            nonspeech_preds = nonspeech_predictor.predict(audio=audio_segment, offset=time_offset)

            if nonspeech_skip is not None:
                segment_nonspeech_timings = nonspeech_preds['timings']

                if segment_nonspeech_timings is not None and len(segment_nonspeech_timings[0]):

                    if (
                            (segment_nonspeech_timings[0][0] <= time_offset + min_word_dur) and
                            (segment_nonspeech_timings[1][0] >= time_offset + segment_samples - min_word_dur)
                    ):
                        seek_sample += segment_samples
                        continue

                    timing_indices = (segment_nonspeech_timings[1] - segment_nonspeech_timings[0]) >= nonspeech_skip
                    if timing_indices.any():
                        nonspeech_starts = segment_nonspeech_timings[0][timing_indices]
                        nonspeech_ends = segment_nonspeech_timings[1][timing_indices]

                        if nonspeech_ends[0] > round(time_offset, 3) >= nonspeech_starts[0]:
                            seek_sample = round(nonspeech_ends[0] * SAMPLE_RATE)
                            if seek_sample + (min_word_dur * SAMPLE_RATE) >= curr_total_samples:
                                seek_sample = curr_total_samples
                                continue
                            time_offset = seek_sample / SAMPLE_RATE

                            audio_segment = audio.next_chunk(seek_sample, N_SAMPLES)
                            if audio_segment is None:
                                break
                            nonspeech_preds = nonspeech_predictor.predict(audio=audio_segment, offset=time_offset)
                            if len(nonspeech_starts) > 1:
                                new_sample_count = round((nonspeech_starts[1] - nonspeech_ends[0]) * SAMPLE_RATE)
                            else:
                                new_sample_count = None
                            audio_segment = audio_segment[:new_sample_count]
                            segment_samples = audio_segment.shape[-1]

            curr_words, curr_word_tokens, curr_split_indices, pad_first_seg = get_curr_words()

            segment = timestamp_words()
            curr_words = segment['words']
            seg_words = [w['word'] for w in curr_words]
            seg_tokens = [w['tokens'] for w in curr_words]
            durations = np.array([w['end'] - w['start'] for w in curr_words]).round(3)
            nonzero_mask = durations > 0
            nonzero_indices = np.flatnonzero(nonzero_mask)
            if len(nonzero_indices):
                redo_index = nonzero_indices[-1] + 1
                if (
                        words and
                        len(nonzero_indices) > 1 and
                        curr_words[nonzero_indices[-1]]['end'] >= np.floor(time_offset + segment_samples / SAMPLE_RATE)
                ):
                    nonzero_mask[nonzero_indices[-1]] = False
                    nonzero_indices = nonzero_indices[:-1]
                    redo_index = nonzero_indices[-1] + 1
                med_dur = np.median(durations[:redo_index])

                if fast_mode:
                    new_start = None
                    global_max_dur = None
                else:
                    local_max_dur = round(med_dur * word_dur_factor, 3) if word_dur_factor else None
                    if max_word_dur:
                        local_max_dur = min(local_max_dur, max_word_dur) if local_max_dur else max_word_dur
                        global_max_dur = max_word_dur
                    else:
                        global_max_dur = local_max_dur or None
                    if global_max_dur and med_dur > global_max_dur:
                        med_dur = global_max_dur
                    if (
                            local_max_dur and durations[nonzero_indices[0]] > global_max_dur
                    ):
                        new_start = round(max(
                            curr_words[nonzero_indices[0]]['end'] - (med_dur * nonzero_indices[0] + local_max_dur),
                            curr_words[nonzero_indices[0]]['start']
                        ), 3)
                        if new_start <= time_offset:
                            new_start = None
                    else:
                        new_start = None
                if new_start is None:
                    if global_max_dur:
                        index_offset = nonzero_indices[0] + 1
                        redo_indices = \
                            np.flatnonzero(durations[index_offset:redo_index] > global_max_dur) + index_offset
                        if len(redo_indices):
                            redo_index = redo_indices[0]
                    last_ts = curr_words[redo_index - 1]['end']
                    redo_words(redo_index)
                else:
                    last_ts = new_start
                    redo_words()
                seek_sample = round(last_ts * SAMPLE_RATE)
            else:
                seek_sample += segment_samples
                last_ts = round(seek_sample / SAMPLE_RATE, 2)
                redo_words()

            update_pbar()

            result.extend(curr_words)

            if verbose:
                line = '\n'.join(
                    f"[{format_timestamp(word['start'])}] -> "
                    f"[{format_timestamp(word['end'])}] \"{word['word']}\""
                    for word in curr_words
                )
                safe_print(line)

            if failure_threshold is not None:
                failure_count += sum(1 for w in curr_words if w['end'] - w['start'] == 0)
                if failure_count > max_fail:
                    break

        update_pbar(failure_count <= max_fail)

    if temp_data.get('temp_word') is not None:
        result.append(temp_data['temp_word'])
    if not result:
        warnings.warn('Failed to align text.', stacklevel=2)
    if failure_count > max_fail:
        warnings.warn(f'Alignment aborted. Failed word percentage exceeded {failure_threshold * 100}% at '
                      f'{format_timestamp(seek_sample / SAMPLE_RATE)}.',
                      stacklevel=2)
    elif words:
        last_ts_str = format_timestamp(result[-1]["end"] if result else 0)
        warnings.warn(f'Failed to align the last {len(words)}/{total_words} words after '
                      f'{last_ts_str}.', stacklevel=2)

    if words and not remove_instant_words:
        final_total_duration = audio.get_duration(3)
        result.extend(
            [
                dict(word=w, start=final_total_duration, end=final_total_duration, probability=0.0, tokens=wt)
                for w, wt in zip(words, word_tokens)
            ]
        )

    audio.terminate()
    nonspeech_predictor.finalize_timings()

    if not result:
        return

    if len(split_indices_by_char):
        word_lens = np.cumsum([[len(w['word']) for w in result]])
        split_indices = [np.flatnonzero(word_lens >= i)[0]+1 for i in split_indices_by_char]
        result = WhisperResult([result[i:j] for i, j in zip([0]+split_indices[:-1], split_indices) if i != j])
    else:
        result = WhisperResult([result])

    if suppress_silence and (nonspeech_timings := nonspeech_predictor.nonspeech_timings) is not None:
        result.suppress_silence(
            *nonspeech_timings,
            min_word_dur=min_word_dur,
            word_level=suppress_word_ts,
            nonspeech_error=nonspeech_error,
            use_word_position=use_word_position,
            verbose=verbose is not None
        )
        result.update_nonspeech_sections(*nonspeech_timings)
    if not original_split:
        result.regroup(regroup)

    if fail_segs := len([None for s in result.segments if s.end-s.start <= 0]):
        warnings.warn(f'{fail_segs}/{len(result.segments)} segments failed to align.', stacklevel=2)

    return result


def refine(
        model: "Whisper",
        audio: Union[str, np.ndarray, torch.Tensor, bytes],
        result: WhisperResult,
        *,
        steps: str = None,
        rel_prob_decrease: float = .03,
        abs_prob_decrease: float = .05,
        rel_rel_prob_decrease: Optional[float] = None,
        prob_threshold: float = .5,
        rel_dur_change: Optional[float] = .5,
        abs_dur_change: Optional[float] = None,
        word_level: bool = True,
        precision: float = None,
        single_batch: bool = False,
        inplace: bool = True,
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: Union[bool, torch.nn.Module] = False,
        demucs_options: dict = None,
        only_voice_freq: bool = False,
        verbose: Optional[bool] = False
) -> WhisperResult:
    """
    Improve existing timestamps.

    This function iteratively muting portions of the audio and monitoring token probabilities to find the most precise
    timestamps. This "most precise" in this case means the latest start and earliest end of a word that maintains an
    acceptable probability determined by the specified arguments.

    This is useful readjusting timestamps when they start too early or end too late.

    Parameters
    ----------
    model : "Whisper"
        The Whisper ASR model modified instance
    audio : str or numpy.ndarray or torch.Tensor or bytes
        Path/URL to the audio file, the audio waveform, or bytes of audio file.
        If audio is :class:`numpy.ndarray` or :class:`torch.Tensor`, the audio must be already at sampled to 16kHz.
    result : stable_whisper.result.WhisperResult
        All timestamps, words, probabilities, and other data from the transcription of ``audio``.
    steps : str, default 'se'
        Instructions for refinement. A 's' means refine start-timestamps. An 'e' means refine end-timestamps.
    rel_prob_decrease : float, default 0.3
        Maximum percent decrease in probability relative to original probability which is the probability from muting
        according initial timestamps.
    abs_prob_decrease : float, default 0.05
        Maximum decrease in probability from original probability.
    rel_rel_prob_decrease : float, optional
        Maximum percent decrease in probability relative to previous probability which is the probability from previous
        iteration of muting.
    prob_threshold : float, default 0.5
        Stop refining the timestamp if the probability of its token goes below this value.
    rel_dur_change : float, default 0.5
        Maximum percent change in duration of a word relative to its original duration.
    abs_dur_change : float, optional
        Maximum seconds a word is allowed deviate from its original duration.
    word_level : bool, default True
        Whether to refine timestamps on word-level. If ``False``, only refine start/end timestamps of each segment.
    precision : float, default 0.1
        Precision of refined timestamps in seconds. The lowest precision is 0.02 second.
    single_batch : bool, default False
        Whether to process in only batch size of one to reduce memory usage.
    inplace : bool, default True, meaning return a deepcopy of ``result``
        Whether to alter timestamps in-place.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.
    verbose : bool or None, default False
        Whether to display the text being decoded to the console.
        Displays all the details if ``True``. Displays progressbar if ``False``. Display nothing if ``None``.

    Returns
    -------
    stable_whisper.result.WhisperResult
        All timestamps, words, probabilities, and other data from the refinement of ``text`` with ``audio``.

    Notes
    -----
    The lower the ``precision``, the longer the processing time.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3')
    >>> model.refine('audio.mp3', result)
    >>> result.to_srt_vtt('audio.srt')
    Saved 'audio.srt'
    """
    audioloader_not_supported(audio)
    if not steps:
        steps = 'se'
    if precision is None:
        precision = 0.1
    if invalid_steps := steps.replace('s', '').replace('e', ''):
        raise ValueError(f'Invalid step(s): {", ".join(invalid_steps)}')
    if not result.has_words:
        raise NotImplementedError(f'Result must have word timestamps.')

    if not inplace:
        result = copy.deepcopy(result)

    audio = prep_audio(
        audio,
        denoiser=denoiser,
        denoiser_options=denoiser_options,
        demucs=demucs,
        demucs_options=demucs_options,
        only_voice_freq=only_voice_freq,
        verbose=verbose
    )
    max_inference_tokens = model.dims.n_text_ctx - 6
    sample_padding = int(N_FFT // 2) + 1
    frame_precision = max(round(precision * FRAMES_PER_SECOND), 2)
    total_duration = round(audio.shape[-1] / SAMPLE_RATE, 3)
    tokenizer = get_tokenizer(model, language=result.language, task='transcribe')

    def ts_to_frames(timestamps: Union[np.ndarray, list]) -> np.ndarray:
        if isinstance(timestamps, list):
            timestamps = np.array(timestamps)
        return (timestamps * FRAMES_PER_SECOND).round().astype(np.int32)

    def curr_segments():
        all_words = result.all_words()
        seg_edge_mask = np.array([
            1 if _i == 0 else (2 if _i == len(seg.words)-1 else 0)
            for seg in result.segments
            for _i, w in enumerate(seg.words)
        ])
        start_times = [
            max(
                0 if abs_dur_change is None else (w.start - abs_dur_change),
                0 if rel_dur_change is None else (w.start - w.duration * rel_dur_change),
                0 if i == 0 else max(all_words[i - 1].end, w.end - 14.5, 0)
            )
            for i, w in enumerate(all_words)
        ]
        end_times = [
            min(
                total_duration if abs_dur_change is None else (w.end + abs_dur_change),
                total_duration if rel_dur_change is None else (w.end + w.duration * rel_dur_change),
                total_duration if i == len(all_words) else min(all_words[i].start, w.start + 14.5, total_duration)
            )
            for i, w in enumerate(all_words, 1)
        ]
        start = start_times[0]

        prev_i = 0
        curr_words, curr_starts, curr_ends = [], [], []
        curr_token_count = 0

        for i, w in enumerate(all_words, 1):
            if (
                    (end_times[0] - start > 30) or
                    (curr_token_count + len(w.tokens) > max_inference_tokens)
            ):
                if curr_words:
                    yield curr_words, curr_starts, curr_ends, seg_edge_mask[prev_i:prev_i+len(curr_words)]
                    curr_words, curr_starts, curr_ends = [], [], []
                start = start_times[0]
                prev_i = i - 1
                curr_token_count = 0

            curr_words.append(w)
            curr_starts.append(start_times.pop(0))
            curr_ends.append(end_times.pop(0))
            curr_token_count += len(w.tokens)

            if i == len(all_words):
                yield curr_words, curr_starts, curr_ends, seg_edge_mask[prev_i:prev_i+len(curr_words)]

    def _refine(_step: str):

        for words, min_starts, max_ends, edge_mask in curr_segments():

            time_offset = min_starts[0]
            start_sample = round(time_offset * SAMPLE_RATE)
            end_sample = round(max_ends[-1] * SAMPLE_RATE)
            audio_segment = audio[start_sample:end_sample + 1].unsqueeze(0)

            max_starts = ts_to_frames(np.array([w.end for w in words]) - time_offset)
            min_ends = ts_to_frames(np.array([w.start for w in words]) - time_offset)
            min_starts = ts_to_frames(np.array(min_starts) - time_offset)
            max_ends = ts_to_frames(np.array(max_ends) - time_offset)

            mid_starts = min_starts + ((max_starts - min_starts) / 2).round().astype(np.int32)
            mid_ends = min_ends + ((max_ends - min_ends) / 2).round().astype(np.int32)

            text_tokens = [t for w in words for t in w.tokens if t < tokenizer.eot]
            word_tokens = [[t for t in w.tokens if t < tokenizer.eot] for w in words]
            orig_mel_segment = log_mel_spectrogram(audio_segment, model.dims.n_mels, padding=sample_padding)
            orig_mel_segment = pad_or_trim(orig_mel_segment, N_FRAMES).to(device=model.device)

            def get_prob():

                tokens = torch.tensor(
                    [
                        *tokenizer.sot_sequence,
                        tokenizer.no_timestamps,
                        *text_tokens,
                        tokenizer.eot,
                    ]
                ).to(model.device)

                with torch.no_grad():
                    curr_mel_segment = mel_segment if prob_indices else orig_mel_segment
                    if single_batch:
                        logits = torch.cat(
                            [model(_mel.unsqueeze(0), tokens.unsqueeze(0)) for _mel in curr_mel_segment]
                        )
                    else:
                        logits = model(curr_mel_segment, tokens.unsqueeze(0))

                sampled_logits = logits[:, len(tokenizer.sot_sequence):, : tokenizer.eot]
                token_probs = sampled_logits.softmax(dim=-1)

                text_token_probs = token_probs[:, np.arange(len(text_tokens)), text_tokens]
                token_positions = token_probs[:, np.arange(len(text_tokens))]
                if logits.shape[0] != 1 and prob_indices is not None:
                    indices1 = np.arange(len(prob_indices))
                    text_token_probs = text_token_probs[prob_indices, indices1]
                    token_positions = token_positions[prob_indices, indices1]
                else:
                    text_token_probs.squeeze_(0)

                text_token_probs = text_token_probs.tolist()
                token_positions = \
                    (
                            token_positions.sort().indices == tokens[len(tokenizer.sot_sequence) + 1:-1][:, None]
                    ).nonzero()[:, -1].tolist()

                word_boundaries = np.pad(np.cumsum([len(t) for t in word_tokens]), (1, 0))
                word_probabilities = np.array([
                    text_token_probs[j-1] if is_end_ts else text_token_probs[i]
                    for i, j in zip(word_boundaries[:-1], word_boundaries[1:])
                ])
                token_positions = [
                    token_positions[j-1] if is_end_ts else token_positions[i]
                    for i, j in zip(word_boundaries[:-1], word_boundaries[1:])
                ]

                return word_probabilities, token_positions

            def update_ts():
                if not is_finish[idx] or changes[idx, -1] == -1:
                    return
                new_ts = round(time_offset + (changes[idx, -1] / FRAMES_PER_SECOND), 3)
                if changes[idx, 0] and not changes[idx, 1]:
                    if is_end_ts:
                        if new_ts <= words[idx].end:
                            return
                    elif new_ts >= words[idx].start:
                        return
                if not verbose:
                    return
                curr_word = words[idx]
                word_info = (f'[Word="{curr_word.word}"] '
                             f'[Segment ID: {curr_word.segment_id}] '
                             f'[Word ID: {curr_word.id}]')
                if is_end_ts:
                    print(f'End: {words[idx].end} -> {new_ts}  {word_info}')
                    words[idx].end = new_ts
                else:
                    print(f'Start: {words[idx].start} -> {new_ts}  {word_info}')
                    words[idx].start = new_ts

            mel_segment = orig_mel_segment.clone().repeat_interleave(2, 0)
            is_end_ts = _step == 'e'

            prob_indices = []
            is_finish = np.less([w.probability for w in words], prob_threshold)
            is_finish = np.logical_or(is_finish, [w.duration == 0 for w in words])
            if not word_level:
                is_finish[edge_mask != (2 if is_end_ts else 1)] = True
            for idx, _i in enumerate(max_starts if is_end_ts else min_ends):
                row = idx % 2
                prob_indices.extend([row] * len(words[idx].tokens))
                if is_finish[idx]:
                    continue
                if is_end_ts:
                    _p = mel_segment.shape[-1] if idx == len(words)-1 else mid_ends[idx+1]
                    mel_segment[row, :, _i:_p] = 0
                else:
                    _p = 0 if idx == 0 else mid_starts[idx-1]
                    mel_segment[row, :, _p:_i] = 0
            orig_probs, orig_tk_poss = get_prob()
            changes = np.zeros((orig_probs.shape[-1], 3), dtype=np.int32)
            changes[:, -1] = -1
            frame_indices = (mid_ends, max_starts) if is_end_ts else (min_ends, mid_starts)
            for idx, (_s, _e) in enumerate(zip(*frame_indices)):
                row = idx % 2
                if is_finish[idx]:
                    continue
                mel_segment[row, :, _s:_e] = 0

            new_probs = prev_probs = orig_probs
            while not np.all(is_finish):
                probs, tk_poss = get_prob()
                abs_diffs = orig_probs - probs
                rel_diffs = abs_diffs / orig_probs
                rel_change_diffs = (prev_probs - probs) / prev_probs
                prev_probs = probs
                for idx, (abs_diff, rel_diff, rel_change_diff, prob) \
                        in enumerate(zip(abs_diffs, rel_diffs, rel_change_diffs, probs)):
                    if is_finish[idx]:
                        continue
                    if is_end_ts:
                        curr_min, curr_max, curr_mid = min_ends[idx], max_ends[idx], mid_ends[idx]
                    else:
                        curr_min, curr_max, curr_mid = min_starts[idx], max_starts[idx], mid_starts[idx]

                    row = prob_indices[idx]
                    best_tks_changed = orig_tk_poss[idx] > tk_poss[idx]
                    failed_requirements = (
                            abs_diff > abs_prob_decrease or
                            rel_diff > rel_prob_decrease or
                            (rel_rel_prob_decrease is not None and rel_change_diff > rel_rel_prob_decrease) or
                            prob < prob_threshold or
                            best_tks_changed
                    )

                    if failed_requirements:
                        changes[idx][0] = 1
                        if is_end_ts:
                            curr_min = curr_mid
                        else:
                            curr_max = curr_mid
                    else:
                        changes[idx][1] = 1
                        if is_end_ts:
                            curr_max = curr_mid
                        else:
                            curr_min = curr_mid

                    if (new_mid_change := round((curr_max - curr_min) / 2)) < frame_precision:
                        is_finish[idx] = True
                        update_ts()
                        continue

                    new_mid = curr_min + new_mid_change
                    if failed_requirements:
                        if is_end_ts:
                            mel_segment[row, :, curr_min:new_mid] = orig_mel_segment[0, :, curr_min:new_mid]
                        else:
                            mel_segment[row, :, new_mid:curr_max] = orig_mel_segment[0, :, new_mid:curr_max]

                    else:
                        if is_end_ts:
                            mel_segment[row, :, new_mid:curr_max] = 0
                        else:
                            mel_segment[row, :, curr_min:new_mid] = 0

                    if is_end_ts:
                        min_ends[idx], max_ends[idx], mid_ends[idx] = curr_min, curr_max, new_mid
                    else:
                        min_starts[idx], max_starts[idx], mid_starts[idx] = curr_min, curr_max, new_mid
                    if not best_tks_changed:
                        changes[idx][-1] = new_mid
                    new_probs[idx] = prob

            update_pbar(words[-1].end)

    with tqdm(total=round(total_duration, 2), unit='sec', disable=verbose is not False, desc='Refine') as tqdm_pbar:

        def update_pbar(last_ts: float):
            nonlocal prev_ts
            tqdm_pbar.update(round(((last_ts - prev_ts) / len(steps)), 2))
            prev_ts = last_ts

        for step_count, step in enumerate(steps, 1):
            prev_ts = 0
            _refine(step)
            update_pbar(round(tqdm_pbar.total / len(step), 2))
        tqdm_pbar.update(tqdm_pbar.total - tqdm_pbar.n)

    result.reassign_ids()

    return result


def locate(
        model: "Whisper",
        audio: Union[str, np.ndarray, torch.Tensor, bytes],
        text: Union[str, List[int]],
        language: str,
        count: int = 1,
        duration_window: Union[float, Tuple[float, float]] = 3.0,
        *,
        mode: int = 0,
        start: float = None,
        end: float = None,
        probability_threshold: float = 0.5,
        eots: int = 1,
        max_token_per_seg: int = 20,
        exact_token: bool = False,
        case_sensitive: bool = False,
        verbose: bool = False,
        initial_prompt: str = None,
        suppress_tokens: Union[str, List[int]] = '-1',
        denoiser: Optional[str] = None,
        denoiser_options: Optional[dict] = None,
        demucs: Union[bool, torch.nn.Module] = False,
        demucs_options: dict = None,
        only_voice_freq: bool = False,
) -> Union[List[Segment], List[dict]]:
    """
    Locate when specific words are spoken in ``audio`` without fully transcribing.

    This is usefully for quickly finding at what time the specify words or phrases are spoken in an audio. Since it
    does not need to transcribe the audio to approximate the time, it is significantly faster transcribing then
    locating the word in the transcript.

    It can also transcribe few seconds around the approximated time to find out what was said around those words or
    confirm if the word was even spoken near that time.

    Parameters
    ----------
    model : whisper.model.Whisper
        An instance of Whisper ASR model.
    audio : str or numpy.ndarray or torch.Tensor or bytes
        Path/URL to the audio file, the audio waveform, or bytes of audio file.
        If audio is :class:`numpy.ndarray` or :class:`torch.Tensor`, the audio must be already at sampled to 16kHz.
    text: str or list of int
        Words/phrase or list of tokens to search for in ``audio``.
    language : str
        Language of the ``text``.
    count : int, default 1, meaning stop search after 1 match
        Number of matches to find. Use 0 to look for all.
    duration_window : float or tuple of (float, float), default 3.0, same as (3.0, 3.0)
        Seconds before and after the end timestamp approximations to transcribe after mode 1.
        If tuple pair of values, then the 1st value will be seconds before the end and 2nd value will be seconds after.
    mode : int, default 0
        Mode of search.
        2, Approximates the end timestamp of ``text`` in the audio. This mode does not confirm whether ``text`` is
            spoken at the timestamp
        1, Completes mode 2 then transcribes audio within ``duration_window`` to confirm whether `text` is a match at
            the approximated timestamp by checking if ``text`` at that ``duration_window`` is within
            ``probability_threshold`` or matching the string content if ``text`` with the transcribed text at the
            ``duration_window``.
        0, Completes mode 1 then add word timestamps to the transcriptions of each match.
        Modes from fastest to slowest: 2, 1, 0
    start : float, optional, meaning it starts from 0s
        Seconds into the audio to start searching for ``text``.
    end : float, optional
        Seconds into the audio to stop searching for ``text``.
    probability_threshold : float, default 0.5
        Minimum probability of each token in ``text`` for it to be considered a match.
    eots : int, default 1
        Number of EOTs to reach before stopping transcription at mode 1. When transcription reach a EOT, it usually
        means the end of the segment or audio. Once ``text`` is found in the ``duration_window``, the transcription
        will stop immediately upon reaching a EOT.
    max_token_per_seg : int, default 20
        Maximum number of tokens to transcribe in the ``duration_window`` before stopping.
    exact_token : bool, default False
        Whether to find a match base on the exact tokens that make up ``text``.
    case_sensitive : bool, default False
        Whether to consider the case of ``text`` when matching in string content.
    verbose : bool or None, default False
        Whether to display the text being decoded to the console.
        Displays all the details if ``True``. Displays progressbar if ``False``. Display nothing if ``None``.
    initial_prompt : str, optional
        Text to provide as a prompt for the first window. This can be used to provide, or
        "prompt-engineer" a context for transcription, e.g. custom vocabularies or proper nouns
        to make it more likely to predict those word correctly.
    suppress_tokens : str or list of int, default '-1', meaning suppress special characters except common punctuations
        List of tokens to suppress.
    denoiser : str, optional
        String of the denoiser to use for preprocessing ``audio``.
        See ``stable_whisper.audio.SUPPORTED_DENOISERS`` for supported denoisers.
    denoiser_options : dict, optional
        Options to use for ``denoiser``.
    only_voice_freq : bool, default False
        Whether to only use sound between 200 - 5000 Hz, where majority of human speech are.

    Returns
    -------
    stable_whisper.result.Segment or list of dict or list of float
        Mode 0, list of instances of :class:`stable_whisper.result.Segment`.
        Mode 1, list of dictionaries with end timestamp approximation of matches and transcribed neighboring words.
        Mode 2, list of timestamps in seconds for each end timestamp approximation.

    Notes
    -----
    For ``text``, the case and spacing matters as 'on', ' on', ' On' are different tokens, therefore chose the one that
    best suits the context (e.g. ' On' to look for it at the beginning of a sentence).

    Use a sufficiently large first value of ``duration_window`` i.e. the value > time it is expected to speak ``text``.

    If ``exact_token = False`` and the string content matches, then ``probability_threshold`` is not used.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> matches = model.locate('audio.mp3', 'are', language='English', verbose=True)

    Some words can sound the same but have different spellings to increase of the chance of finding such words use
    ``initial_prompt``.

    >>> matches = model.locate('audio.mp3', ' Nickie', 'English', verbose=True, initial_prompt='Nickie')
    """
    audioloader_not_supported(audio)

    sample_padding = int(N_FFT // 2) + 1
    sec_per_emb = model.dims.n_audio_ctx / CHUNK_LENGTH
    CHUNK_SAMPLES = round(CHUNK_LENGTH * SAMPLE_RATE)
    if isinstance(duration_window, (float, int)):
        duration_window = [duration_window] * 2
    window_sum = sum(duration_window)
    assert CHUNK_SAMPLES > window_sum, \
        f'Sum of [duration_window] must be less than {CHUNK_SAMPLES}, got {window_sum}'
    adjusted_chunk_size = CHUNK_SAMPLES - round(duration_window[0]*SAMPLE_RATE)
    if initial_prompt:
        initial_prompt = ' ' + initial_prompt.strip()
    task = DecodingTask(model, DecodingOptions(
        language=language, prompt=initial_prompt, suppress_tokens=suppress_tokens, without_timestamps=True,
    ))
    tokenizer = task.tokenizer
    initial_tokens = list(task.initial_tokens)
    text_tokens, text = (tokenizer.encode(text), text) if isinstance(text, str) else (text, tokenizer.decode(text))
    if not exact_token and not case_sensitive:
        text = text.lower()

    tk_suppress_masks = [
        [i for i in fil.suppress_tokens if i < tokenizer.eot]
        for fil in task.logit_filters if isinstance(fil, SuppressTokens)
    ]

    audio = prep_audio(
        audio,
        denoiser=denoiser,
        denoiser_options=denoiser_options,
        demucs=demucs,
        demucs_options=demucs_options,
        only_voice_freq=only_voice_freq,
        verbose=verbose
    )
    prev_target_end = None
    found = 0
    if end:
        audio = audio[:round(end * SAMPLE_RATE)]
    seek_sample = round(start * SAMPLE_RATE) if start else 0
    total_samples = audio.shape[-1]

    def _locate():
        nonlocal seek_sample, found
        seek = round(seek_sample / SAMPLE_RATE, 3)
        audio_segment = audio[seek_sample: seek_sample + CHUNK_SAMPLES]
        mel_segment = log_mel_spectrogram(audio_segment, model.dims.n_mels, padding=sample_padding)
        mel_segment = pad_or_trim(mel_segment, N_FRAMES).to(device=model.device)

        QKs = [None] * model.dims.n_text_layer
        hooks = [
            block.cross_attn.register_forward_hook(
                lambda _, ins, outs, index=i: QKs.__setitem__(index, outs[-1])
            )
            for i, block in enumerate(model.decoder.blocks)
        ]
        tokens = torch.tensor([initial_tokens + text_tokens]).to(model.device)
        with torch.no_grad():
            audio_features = model.encoder(mel_segment.unsqueeze(0))
            model.decoder(tokens, audio_features)

        for hook in hooks:
            hook.remove()

        weights = torch.cat([QKs[_l][:, _h] for _l, _h in model.alignment_heads.indices().T], dim=0)
        weights = weights.softmax(dim=-1)
        std, mean = torch.std_mean(weights, dim=-2, keepdim=True, unbiased=False)
        weights = (weights - mean) / std
        weights = median_filter(weights, 7)

        matrix = weights.mean(axis=0)
        target_end = round((matrix[-1].argmax()/sec_per_emb).item(), 3)
        found_msg = f'"{text}" ending at ~{format_timestamp(target_end+seek)}' if verbose else ''

        if mode == 2:
            if found_msg:
                safe_print('Unconfirmed:' + found_msg)
            nonlocal prev_target_end
            found += 1
            if (
                    (seek_sample + CHUNK_SAMPLES >= total_samples) or
                    (count and found >= count) or
                    (prev_target_end == target_end)
            ):
                seek_sample = total_samples
            else:
                seek_sample += round(target_end * SAMPLE_RATE)
            prev_target_end = target_end
            return dict(tokens=[], target_end=target_end+seek)

        curr_start = round(max(target_end - duration_window[0], 0.), 3)
        curr_end = round(target_end + duration_window[1], 3)
        start_frame = round(curr_start * FRAMES_PER_SECOND)
        end_frame = round(curr_end * FRAMES_PER_SECOND)
        mel_segment_section = pad_or_trim(mel_segment[..., start_frame:end_frame], N_FRAMES)
        temp_tokens = torch.tensor([initial_tokens]).to(model.device)

        predictions = []

        target_token_idx = 0
        not_end = True
        found_target = False
        curr_eots = 0
        temp_audio_features = model.encoder(mel_segment_section.unsqueeze(0))
        tokens_to_decode = []
        replace_found_tokens = []
        infer_tokens = [temp_tokens[0]]
        kv_cache, hooks = model.install_kv_cache_hooks()
        while not_end:
            with torch.no_grad():
                logits = model.decoder(temp_tokens, temp_audio_features, kv_cache=kv_cache)[0, -1, :tokenizer.eot+1]
            for tks in tk_suppress_masks:
                logits[tks] = -np.inf
            sorted_logits_idxs = logits.sort(dim=-1).indices[-2:]
            best_token = sorted_logits_idxs[-1]
            best_non_eot_token = sorted_logits_idxs[-2] if best_token == tokenizer.eot else best_token

            logits = logits[:tokenizer.eot].softmax(dim=-1)
            if found_target:
                target_word_prob = is_match = None
            else:
                if exact_token:
                    is_match = False
                else:
                    tokens_to_decode.append(best_non_eot_token)
                    temp_text = tokenizer.decode(tokens_to_decode)
                    if not case_sensitive:
                        temp_text = temp_text.lower()
                    if is_match := temp_text.endswith(text):
                        tokens_to_decode = []
                target_word_prob = logits[text_tokens[target_token_idx]].item()
            if (
                    target_word_prob is not None and
                    (
                            target_word_prob >= probability_threshold or
                            best_non_eot_token == text_tokens[target_token_idx] or
                            is_match
                    )
            ):
                if is_match:
                    best_token = best_non_eot_token
                    token_prob = logits[best_token].item()
                    found_target = True
                else:
                    best_token[None] = text_tokens[target_token_idx]
                    if len(replace_found_tokens) or best_non_eot_token != text_tokens[target_token_idx]:
                        replace_found_tokens.append(best_non_eot_token)
                    target_token_idx += 1
                    if target_token_idx == len(text_tokens):
                        found_target = True
                    token_prob = target_word_prob
                if found_target:
                    found += 1
                curr_eots = 0
            else:
                if not found_target:
                    if len(replace_found_tokens):
                        temp_tokens = torch.cat(infer_tokens)[None]
                        temp_tokens = torch.cat(
                            [temp_tokens[..., :-len(replace_found_tokens)],
                             torch.stack(replace_found_tokens)[None]]
                        )
                        replace_found_tokens = []
                        kv_cache.clear()
                    target_token_idx = 0
                if best_token == tokenizer.eot:
                    if curr_eots >= eots or found_target:
                        not_end = False
                    else:
                        curr_eots += 1
                        best_token = best_non_eot_token
                else:
                    curr_eots = 0
                token_prob = None if best_token == tokenizer.eot else logits[best_token].item()

            predictions.append(dict(token=best_token.item(), prob=token_prob))
            if len(predictions) > max_token_per_seg:
                not_end = False
            if not_end:
                infer_tokens.append(best_token[None])
                temp_tokens = best_token[None, None]
        kv_cache.clear()
        for hook in hooks:
            hook.remove()
        segment = None

        if found_target:
            if found_msg:
                safe_print('Confirmed: ' + found_msg, tqdm_pbar.write)
            final_tokens = [p['token'] for p in predictions]
            if mode == 1:
                _, (ws, wts), _ = split_word_tokens([dict(tokens=final_tokens)], tokenizer)
                final_token_probs = [p['prob'] for p in predictions]
                wps = [float(np.mean([final_token_probs.pop(0) for _ in wt])) for wt in wts]
                words = [dict(word=w, tokens=wt, probability=wp) for w, wt, wp in zip(ws, wts, wps)]
                final_end = target_end+seek
                near_text = "".join(ws)
                segment = dict(end=final_end, text=text, duration_window_text=near_text, duration_window_word=words)
                if verbose:
                    safe_print(f'Duration Window: "{near_text}"\n', tqdm_pbar.write)
                seek_sample += round(curr_end * SAMPLE_RATE)
            else:

                segment = dict(
                    seek=0,
                    tokens=final_tokens
                )

                add_word_timestamps_stable(
                    segments=[segment],
                    model=model,
                    tokenizer=tokenizer,
                    mel=mel_segment,
                    num_samples=round(curr_end*SAMPLE_RATE),
                    gap_padding=None
                )
                segment = Segment(words=segment['words'])
                seek_sample += round(segment.words[-1].end * SAMPLE_RATE)
                segment.offset_time(seek)
                segment.seek = curr_start
                if verbose:
                    safe_print(segment.to_display_str(), tqdm_pbar.write)

        else:
            seek_sample += adjusted_chunk_size if audio_segment.shape[-1] == CHUNK_SAMPLES else audio_segment.shape[-1]

        return segment

    total_duration = round(total_samples / SAMPLE_RATE, 2)
    matches = []
    with tqdm(total=total_duration, unit='sec', disable=verbose is None, desc='Locate') as tqdm_pbar:
        while seek_sample < total_samples and (not count or found < count):
            if match := _locate():
                matches.append(match)
            tqdm_pbar.update(round(seek_sample/SAMPLE_RATE, 2) - tqdm_pbar.n)
        tqdm_pbar.update(tqdm_pbar.total - tqdm_pbar.n)
    if verbose and not matches:
        safe_print(f'Failed to locate "{text}".')
    return matches



from .whisper_word_level import *
from .result import *
from .text_output import *
from .video_output import *
from .stabilization import visualize_suppression
from .non_whisper import transcribe_any
from ._version import __version__
from .whisper_compatibility import _required_whisper_ver, _COMPATIBLE_WHISPER_VERSIONS, IS_WHISPERLESS_VERSION



import os
from typing import Optional

DEFAULT_KWARGS = dict(
    prepend_punctuations="\"'“¿([{-「",
    append_punctuations="\"'.。,，!！?？:：”)]}、」",
    min_word_dur=0.1
)

permissions = {}

cached_model_instances = dict(
    demucs={
        'htdemucs': None
    },
    silero_vad={
        True: None,
        False: None
    },
    dfnet={
        'dfnet': None
    }
)

__all__ = [
    'get_prepend_punctuations',
    'get_append_punctuations',
    'get_min_word_dur',
    'is_allow_overwrite',
    'set_global_overwrite_permission',
    'cached_model_instances'
]


def has_key(key: str):
    if key not in DEFAULT_KWARGS:
        raise KeyError(f'the key, {key}, not found in DEFAULT_VALUES')
    return True


def set_val(key: str, val):
    has_key(key)
    DEFAULT_KWARGS[key] = val


def get_val(key: str, default=None):
    if default is None:
        has_key(key)
        return DEFAULT_KWARGS[key]
    return default


def set_get_val(key: str, new_val=None):
    if new_val is not None:
        set_val(key, new_val)
    return get_val(key)


def get_prepend_punctuations(default: Optional[str] = None) -> str:
    return get_val('prepend_punctuations', default)


def get_append_punctuations(default: Optional[str] = None) -> str:
    return get_val('append_punctuations', default)


def get_min_word_dur(default: Optional[float] = None) -> float:
    return get_val('min_word_dur', default)


def is_allow_overwrite(filepath: str, default: (bool, None) = None) -> bool:
    if default is not None:
        return default
    if not os.path.isfile(filepath) or permissions.get('overwrite'):
        return True
    resp = input(f'"{filepath}" already exist, overwrite (y/n)? ').lower()
    if resp in ('y', 'n'):
        return resp == 'y'
    print(f'Expected "y" or "n", but got {resp}.')
    return is_allow_overwrite(filepath)


def set_global_overwrite_permission(overwrite: bool):
    permissions['overwrite'] = overwrite



from typing import Union, Optional

import torch

from .utils import load_audio
from ..audio.utils import resample
from ..default import cached_model_instances


def is_dfnet_available():
    from importlib.util import find_spec
    if find_spec('df') is None:
        raise ModuleNotFoundError("Please install DeepFilterNet; "
                                  "'pip install -U deepfilternet' or "
                                  "follow installation instructions at https://github.com/Rikorose/DeepFilterNet")


def load_dfnet_model(cache: bool = True, **kwargs):
    model_name = 'dfnet'
    _model_cache = cached_model_instances['dfnet'] if cache else None
    if _model_cache is not None and _model_cache[model_name] is not None:
        return _model_cache[model_name]
    is_dfnet_available()
    from types import MethodType
    from df.enhance import init_df, enhance
    model, df_state, _ = init_df(**kwargs)
    model.df_state = df_state

    def enhance_wrapper(_model, audio, **enhance_kwargs):
        return enhance(model=_model, df_state=_model.df_state, audio=audio, **enhance_kwargs)

    model.enhance = MethodType(enhance_wrapper, model)
    model.samplerate = df_state.sr()
    if _model_cache is not None:
        _model_cache[model_name] = model
    return model


def dfnet_audio(
        audio: Union[torch.Tensor, str, bytes],
        input_sr: int = None,
        output_sr: int = None,
        model=None,
        device=None,
        verbose: bool = True,
        save_path: Optional[Union[str, callable]] = None,
        **dfnet_options
) -> torch.Tensor:
    """
    Remove noise from ``audio`` with DeepFilterNet.

    Official repo: https://github.com/Rikorose/DeepFilterNet.
    """
    if model is None:
        model = load_dfnet_model()
    if isinstance(audio, (str, bytes)):
        audio = torch.from_numpy(load_audio(audio, model.samplerate, mono=False))
    elif input_sr != model.samplerate:
        if input_sr is None:
            raise ValueError('No ``input_sr`` specified for audio tensor.')
        audio = resample(audio, input_sr, model.samplerate)
    audio_dims = audio.dim()
    assert audio_dims <= 2
    if dims_missing := 2 - audio_dims:
        audio = audio[[None]*dims_missing]
    if audio.shape[-2] == 1:
        audio = audio.repeat_interleave(2, -2)

    if device is not None:
        from df import config
        device_str = str(device)
        config.set('DEVICE', device_str, str, section='train')
        audio.to(device=device)
        model.to(device=device)

    dfnet_options.pop('progress', None)  # not implemented
    denoised_audio = model.enhance(audio=audio, **dfnet_options).mean(dim=0)

    if 'cuda' in str(device):
        torch.cuda.empty_cache()

    if output_sr is not None and model.samplerate != output_sr:
        denoised_audio = resample(denoised_audio, model.samplerate, output_sr)

    if save_path is not None:
        if isinstance(save_path, str):
            from .output import save_audio_tensor
            save_audio_tensor(denoised_audio, save_path, output_sr or model.samplerate, verbose=verbose)
        else:
            save_path(denoised_audio)

    return denoised_audio



import json
import os
import warnings
from typing import List, Tuple, Union, Callable
from itertools import chain
from .stabilization.utils import valid_ts

__all__ = ['result_to_srt_vtt', 'result_to_ass', 'result_to_tsv', 'result_to_txt', 'save_as_json', 'load_result']
SUPPORTED_FORMATS = ('srt', 'vtt', 'ass', 'tsv', 'txt')


def _save_as_file(content: str, path: str):
    with open(path, 'w', encoding='utf-8') as f:
        f.write(content)
    print(f'Saved: {os.path.abspath(path)}')


def _get_segments(result: (dict, list), min_dur: float, reverse_text: Union[bool, tuple] = False):
    if isinstance(result, dict):
        if reverse_text:
            warnings.warn(f'``reverse_text=True`` only applies to WhisperResult but result is {type(result)}')
        return result.get('segments')
    elif not isinstance(result, list) and callable(getattr(result, 'segments_to_dicts', None)):
        return result.apply_min_dur(min_dur, inplace=False).segments_to_dicts(reverse_text=reverse_text)
    return result


def finalize_text(text: str, strip: bool = True):
    if not strip:
        return text
    return text.strip().replace('\n ', '\n')


def sec2hhmmss(seconds: (float, int)):
    mm, ss = divmod(seconds, 60)
    hh, mm = divmod(mm, 60)
    return hh, mm, ss


def sec2milliseconds(seconds: (float, int)) -> int:
    return round(seconds * 1000)


def sec2centiseconds(seconds: (float, int)) -> int:
    return round(seconds * 100)


def sec2vtt(seconds: (float, int)) -> str:
    hh, mm, ss = sec2hhmmss(seconds)
    return f'{hh:0>2.0f}:{mm:0>2.0f}:{ss:0>6.3f}'


def sec2srt(seconds: (float, int)) -> str:
    return sec2vtt(seconds).replace(".", ",")


def sec2ass(seconds: (float, int)) -> str:
    hh, mm, ss = sec2hhmmss(seconds)
    return f'{hh:0>1.0f}:{mm:0>2.0f}:{ss:0>2.2f}'


def segment2vttblock(segment: dict, strip=True) -> str:
    return f'{sec2vtt(segment["start"])} --> {sec2vtt(segment["end"])}\n' \
           f'{finalize_text(segment["text"], strip)}'


def segment2srtblock(segment: dict, idx: int, strip=True) -> str:
    return f'{idx}\n{sec2srt(segment["start"])} --> {sec2srt(segment["end"])}\n' \
           f'{finalize_text(segment["text"], strip)}'


def segment2assblock(segment: dict, idx: int, strip=True) -> str:
    return f'Dialogue: {idx},{sec2ass(segment["start"])},{sec2ass(segment["end"])},Default,,0,0,0,,' \
           f'{finalize_text(segment["text"], strip)}'


def segment2tsvblock(segment: dict, strip=True) -> str:
    return f'{sec2milliseconds(segment["start"])}' \
           f'\t{sec2milliseconds(segment["end"])}' \
           f'\t{segment["text"].strip() if strip else segment["text"]}'


def words2segments(words: List[dict], tag: Tuple[str, str], reverse_text: bool = False) -> List[dict]:
    def add_tag(idx: int):
        return ''.join(
            (
                f" {tag[0]}{w['word'][1:]}{tag[1]}"
                if w['word'].startswith(' ') else
                f"{tag[0]}{w['word']}{tag[1]}"
            )
            if w['word'] not in ('', ' ') and idx_ == idx else
            w['word']
            for idx_, w in idx_filled_words
        )

    filled_words = []
    for i, word in enumerate(words):
        curr_end = round(word['end'], 3)
        filled_words.append(dict(word=word['word'], start=round(word['start'], 3), end=curr_end))
        if word != words[-1]:
            next_start = round(words[i + 1]['start'], 3)
            if next_start - curr_end != 0:
                filled_words.append(dict(word='', start=curr_end, end=next_start))
    idx_filled_words = list(enumerate(filled_words))
    if reverse_text:
        idx_filled_words = list(reversed(idx_filled_words))

    segments = [dict(text=add_tag(i), start=filled_words[i]['start'], end=filled_words[i]['end'])
                for i in range(len(filled_words))]
    return segments


def to_word_level_segments(segments: List[dict], tag: Tuple[str, str]) -> List[dict]:
    return list(
        chain.from_iterable(
            words2segments(s['words'], tag, reverse_text=s.get('reversed_text'))
            for s in segments
        )
    )


def to_vtt_word_level_segments(segments: List[dict], tag: Tuple[str, str] = None) -> List[dict]:
    def to_segment_string(segment: dict):
        segment_string = ''
        prev_end = 0
        for i, word in enumerate(segment['words']):
            if i != 0:
                curr_start = word['start']
                if prev_end == curr_start:
                    segment_string += f"<{sec2vtt(curr_start)}>"
                else:
                    if segment_string.endswith(' '):
                        segment_string = segment_string[:-1]
                    elif segment['words'][i]['word'].startswith(' '):
                        segment['words'][i]['word'] = segment['words'][i]['word'][1:]
                    segment_string += f"<{sec2vtt(prev_end)}> <{sec2vtt(curr_start)}>"
            segment_string += word['word']
            prev_end = word['end']
        return segment_string

    return [
        dict(
            text=to_segment_string(s),
            start=s['start'],
            end=s['end']
        )
        for s in segments
    ]


def to_ass_word_level_segments(segments: List[dict], tag: Tuple[str, str], karaoke: bool = False) -> List[dict]:

    def to_segment_string(segment: dict):
        segment_string = ''
        for i, word in enumerate(segment['words']):
            curr_word, space = (word['word'][1:], " ") if word['word'].startswith(" ") else (word['word'], "")
            segment_string += (
                    space +
                    r"{\k" +
                    ("f" if karaoke else "") +
                    f"{sec2centiseconds(word['end']-word['start'])}" +
                    r"}" +
                    curr_word
            )
        return segment_string

    return [
        dict(
            text=to_segment_string(s),
            start=s['start'],
            end=s['end']
        )
        for s in segments
    ]


def to_word_level(segments: List[dict]) -> List[dict]:
    return [dict(text=w['word'], start=w['start'], end=w['end']) for s in segments for w in s['words']]


def _confirm_word_level(segments: List[dict]) -> bool:
    if not all(bool(s.get('words')) for s in segments):
        warnings.warn('Result is missing word timestamps. Word-level timing cannot be exported. '
                      'Use ``word_level=False`` to avoid this warning')
        return False
    return True


def _preprocess_args(result: (dict, list),
                     segment_level: bool,
                     word_level: bool,
                     min_dur: float,
                     reverse_text: Union[bool, tuple] = False):
    assert segment_level or word_level, '`segment_level` or `word_level` must be True'
    segments = _get_segments(result, min_dur, reverse_text=reverse_text)
    if word_level:
        word_level = _confirm_word_level(segments)
    return segments, segment_level, word_level


def result_to_any(result: (dict, list),
                  filepath: str = None,
                  filetype: str = None,
                  segments2blocks: Callable = None,
                  segment_level=True,
                  word_level=True,
                  min_dur: float = 0.02,
                  tag: Tuple[str, str] = None,
                  default_tag: Tuple[str, str] = None,
                  strip=True,
                  reverse_text: Union[bool, tuple] = False,
                  to_word_level_string_callback: Callable = None):
    """
    Generate file from ``result`` to display segment-level and/or word-level timestamp.

    Returns
    -------
    str
        String of the content if ``filepath`` is ``None``.
    """
    segments, segment_level, word_level = _preprocess_args(
        result, segment_level, word_level, min_dur, reverse_text=reverse_text
    )

    if filetype is None:
        filetype = os.path.splitext(filepath)[-1][1:] or 'srt'
    if filetype.lower() not in SUPPORTED_FORMATS:
        raise NotImplementedError(f'{filetype} not supported')
    if filepath and not filepath.lower().endswith(f'.{filetype}'):
        filepath += f'.{filetype}'

    if word_level and segment_level:
        if tag is None:
            if default_tag is None:
                tag = ('<font color="#00ff00">', '</font>') if filetype == 'srt' else ('<u>', '</u>')
            else:
                tag = default_tag
        if to_word_level_string_callback is None:
            to_word_level_string_callback = to_word_level_segments
        segments = to_word_level_string_callback(segments, tag)
    elif word_level:
        segments = to_word_level(segments)

    if not valid_ts(segments, warn=False):
        warnings.warn(message='Result contains out of order timestamp(s). Output file may not playback properly.')

    if segments2blocks is None:
        sub_str = '\n\n'.join(segment2srtblock(s, i, strip=strip) for i, s in enumerate(segments))
    else:
        sub_str = segments2blocks(segments)

    if filepath:
        _save_as_file(sub_str, filepath)
    else:
        return sub_str


def result_to_srt_vtt(result: (dict, list),
                      filepath: str = None,
                      segment_level=True,
                      word_level=True,
                      min_dur: float = 0.02,
                      tag: Tuple[str, str] = None,
                      vtt: bool = None,
                      strip=True,
                      reverse_text: Union[bool, tuple] = False):
    """
    Generate SRT/VTT from ``result`` to display segment-level and/or word-level timestamp.

    Parameters
    ----------
    result : dict or list or stable_whisper.result.WhisperResult
        Result of transcription.
    filepath : str, default None, meaning content will be returned as a ``str``
        Path to save file.
    segment_level : bool, default True
        Whether to use segment-level timestamps in output.
    word_level : bool, default True
        Whether to use word-level timestamps in output.
    min_dur : float, default 0.2
        Minimum duration allowed for any word/segment before the word/segments are merged with adjacent word/segments.
    tag: tuple of (str, str), default None, meaning ('<font color="#00ff00">', '</font>') if SRT else ('<u>', '</u>')
        Tag used to change the properties a word at its timestamp.
    vtt : bool, default None, meaning determined by extension of ``filepath`` or ``False`` if no valid extension.
        Whether to output VTT.
    strip : bool, default True
        Whether to remove spaces before and after text on each segment for output.
    reverse_text: bool or tuple, default False
        Whether to reverse the order of words for each segment or provide the ``prepend_punctuations`` and
        ``append_punctuations`` as tuple pair instead of ``True`` which is for the default punctuations.

    Returns
    -------
    str
        String of the content if ``filepath`` is ``None``.

    Notes
    -----
    ``reverse_text`` will not fix RTL text not displaying tags properly which is an issue with some video player. VLC
    seems to not suffer from this issue.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3')
    >>> result.to_srt_vtt('audio.srt')
    Saved: audio.srt
    """
    is_srt = (filepath is None or not filepath.lower().endswith('.vtt')) if vtt is None else not vtt
    if is_srt:
        segments2blocks = None
        to_word_level_string_callback = None
    else:
        def segments2blocks(segments):
            return 'WEBVTT\n\n' + '\n\n'.join(segment2vttblock(s, strip=strip) for i, s in enumerate(segments))
        to_word_level_string_callback = to_vtt_word_level_segments if tag is None else tag

    return result_to_any(
        result=result,
        filepath=filepath,
        filetype=('vtt', 'srt')[is_srt],
        segments2blocks=segments2blocks,
        segment_level=segment_level,
        word_level=word_level,
        min_dur=min_dur,
        tag=tag,
        strip=strip,
        reverse_text=reverse_text,
        to_word_level_string_callback=to_word_level_string_callback
    )


def result_to_tsv(result: (dict, list),
                  filepath: str = None,
                  segment_level: bool = None,
                  word_level: bool = None,
                  min_dur: float = 0.02,
                  strip=True,
                  reverse_text: Union[bool, tuple] = False):
    """
    Generate TSV from ``result`` to display segment-level and/or word-level timestamp.

    Parameters
    ----------
    result : dict or list or stable_whisper.result.WhisperResult
        Result of transcription.
    filepath : str, default None, meaning content will be returned as a ``str``
        Path to save file.
    segment_level : bool, default True
        Whether to use segment-level timestamps in output.
    word_level : bool, default True
        Whether to use word-level timestamps in output.
    min_dur : float, default 0.2
        Minimum duration allowed for any word/segment before the word/segments are merged with adjacent word/segments.
    strip : bool, default True
        Whether to remove spaces before and after text on each segment for output.
    reverse_text: bool or tuple, default False
        Whether to reverse the order of words for each segment or provide the ``prepend_punctuations`` and
        ``append_punctuations`` as tuple pair instead of ``True`` which is for the default punctuations.

    Returns
    -------
    str
        String of the content if ``filepath`` is ``None``.

    Notes
    -----
    ``reverse_text`` will not fix RTL text not displaying tags properly which is an issue with some video player. VLC
    seems to not suffer from this issue.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3')
    >>> result.to_tsv('audio.tsv')
    Saved: audio.tsv
    """
    if segment_level is None and word_level is None:
        segment_level = True
    assert word_level is not segment_level, '[word_level] and [segment_level] cannot be the same ' \
                                            'since [tag] is not support for this format'

    def segments2blocks(segments):
        return '\n\n'.join(segment2tsvblock(s, strip=strip) for i, s in enumerate(segments))
    return result_to_any(
        result=result,
        filepath=filepath,
        filetype='tsv',
        segments2blocks=segments2blocks,
        segment_level=segment_level,
        word_level=word_level,
        min_dur=min_dur,
        strip=strip,
        reverse_text=reverse_text
    )


def result_to_ass(result: (dict, list),
                  filepath: str = None,
                  segment_level=True,
                  word_level=True,
                  min_dur: float = 0.02,
                  tag: Union[Tuple[str, str], int] = None,
                  font: str = None,
                  font_size: int = 24,
                  strip=True,
                  highlight_color: str = None,
                  karaoke=False,
                  reverse_text: Union[bool, tuple] = False,
                  **kwargs):
    """
    Generate Advanced SubStation Alpha (ASS) file from ``result`` to display segment-level and/or word-level timestamp.

    Parameters
    ----------
    result : dict or list or stable_whisper.result.WhisperResult
        Result of transcription.
    filepath : str, default None, meaning content will be returned as a ``str``
        Path to save file.
    segment_level : bool, default True
        Whether to use segment-level timestamps in output.
    word_level : bool, default True
        Whether to use word-level timestamps in output.
    min_dur : float, default 0.2
        Minimum duration allowed for any word/segment before the word/segments are merged with adjacent word/segments.
    tag: tuple of (str, str) or int, default None, meaning use default highlighting
        Tag used to change the properties a word at its timestamp. -1 for individual word highlight tag.
    font : str, default `Arial`
        Word font.
    font_size : int, default 48
        Word font size.
    strip : bool, default True
        Whether to remove spaces before and after text on each segment for output.
    highlight_color : str, default '00ff00'
        Hexadecimal of the color use for default highlights as '<bb><gg><rr>'.
    karaoke : bool, default False
        Whether to use progressive filling highlights (for karaoke effect).
    reverse_text: bool or tuple, default False
        Whether to reverse the order of words for each segment or provide the ``prepend_punctuations`` and
        ``append_punctuations`` as tuple pair instead of ``True`` which is for the default punctuations.
    kwargs:
        Format styles:
        'Name', 'Fontname', 'Fontsize', 'PrimaryColour', 'SecondaryColour', 'OutlineColour', 'BackColour', 'Bold',
        'Italic', 'Underline', 'StrikeOut', 'ScaleX', 'ScaleY', 'Spacing', 'Angle', 'BorderStyle', 'Outline',
        'Shadow', 'Alignment', 'MarginL', 'MarginR', 'MarginV', 'Encoding'

    Returns
    -------
    str
        String of the content if ``filepath`` is ``None``.

    Notes
    -----
    ``reverse_text`` will not fix RTL text not displaying tags properly which is an issue with some video player. VLC
    seems to not suffer from this issue.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3')
    >>> result.to_ass('audio.ass')
    Saved: audio.ass
    """
    if tag == ['-1']:  # CLI
        tag = -1
    if highlight_color is None:
        highlight_color = '00ff00'

    def segments2blocks(segments):
        fmt_style_dict = {'Name': 'Default', 'Fontname': 'Arial', 'Fontsize': '48', 'PrimaryColour': '&Hffffff',
                          'SecondaryColour': '&Hffffff', 'OutlineColour': '&H0', 'BackColour': '&H0', 'Bold': '0',
                          'Italic': '0', 'Underline': '0', 'StrikeOut': '0', 'ScaleX': '100', 'ScaleY': '100',
                          'Spacing': '0', 'Angle': '0', 'BorderStyle': '1', 'Outline': '1', 'Shadow': '0',
                          'Alignment': '2', 'MarginL': '10', 'MarginR': '10', 'MarginV': '10', 'Encoding': '0'}

        for k, v in filter(lambda x: 'colour' in x[0].lower() and not str(x[1]).startswith('&H'), kwargs.items()):
            kwargs[k] = f'&H{kwargs[k]}'

        fmt_style_dict.update((k, v) for k, v in kwargs.items() if k in fmt_style_dict)

        if tag is None and 'PrimaryColour' not in kwargs:
            fmt_style_dict['PrimaryColour'] = \
                highlight_color if highlight_color.startswith('&H') else f'&H{highlight_color}'

        if font:
            fmt_style_dict.update(Fontname=font)
        if font_size:
            fmt_style_dict.update(Fontsize=font_size)

        fmts = f'Format: {", ".join(map(str, fmt_style_dict.keys()))}'

        styles = f'Style: {",".join(map(str, fmt_style_dict.values()))}'

        sub_str = f'[Script Info]\nScriptType: v4.00+\nPlayResX: 384\nPlayResY: 288\nScaledBorderAndShadow: yes\n\n' \
                  f'[V4+ Styles]\n{fmts}\n{styles}\n\n' \
                  f'[Events]\nFormat: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text\n\n'

        sub_str += '\n'.join(segment2assblock(s, i, strip=strip) for i, s in enumerate(segments))

        return sub_str

    if tag is not None and karaoke:
        warnings.warn(f'``tag`` is not support for ``karaoke=True``; ``tag`` will be ignored.')

    return result_to_any(
        result=result,
        filepath=filepath,
        filetype='ass',
        segments2blocks=segments2blocks,
        segment_level=segment_level,
        word_level=word_level,
        min_dur=min_dur,
        tag=None if tag == -1 else tag,
        default_tag=(r'{\1c' + f'{highlight_color}&' + '}', r'{\r}'),
        strip=strip,
        reverse_text=reverse_text,
        to_word_level_string_callback=(
            (lambda s, t: to_ass_word_level_segments(s, t, karaoke=karaoke))
            if karaoke or (word_level and segment_level and tag is None)
            else None
        )
    )


def result_to_txt(
        result: (dict, list),
        filepath: str = None,
        min_dur: float = 0.02,
        strip=True,
        reverse_text: Union[bool, tuple] = False
):
    """
    Generate plain-text without timestamps from ``result``.

    Parameters
    ----------
    result : dict or list or stable_whisper.result.WhisperResult
        Result of transcription.
    filepath : str, default None, meaning content will be returned as a ``str``
        Path to save file.
    min_dur : float, default 0.2
        Minimum duration allowed for any word/segment before the word/segments are merged with adjacent word/segments.
    strip : bool, default True
        Whether to remove spaces before and after text on each segment for output.
    reverse_text: bool or tuple, default False
        Whether to reverse the order of words for each segment or provide the ``prepend_punctuations`` and
        ``append_punctuations`` as tuple pair instead of ``True`` which is for the default punctuations.

    Returns
    -------
    str
        String of the content if ``filepath`` is ``None``.

    Notes
    -----
    ``reverse_text`` will not fix RTL text not displaying tags properly which is an issue with some video player. VLC
    seems to not suffer from this issue.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3')
    >>> result.to_txt('audio.txt')
    Saved: audio.txt
    """

    def segments2blocks(segments: dict, _strip=True) -> str:
        return '\n'.join(f'{segment["text"].strip() if _strip else segment["text"]}' for segment in segments)

    return result_to_any(
        result=result,
        filepath=filepath,
        filetype='txt',
        segments2blocks=segments2blocks,
        segment_level=True,
        word_level=False,
        min_dur=min_dur,
        strip=strip,
        reverse_text=reverse_text
    )


def save_as_json(result: dict, path: str, ensure_ascii: bool = False, **kwargs):
    """
    Save ``result`` as JSON file to ``path``.

    Parameters
    ----------
    result : dict or list or stable_whisper.result.WhisperResult
        Result of transcription.
    path : str
        Path to save file.
    ensure_ascii : bool, default False
        Whether to escape non-ASCII characters.

    Examples
    --------
    >>> import stable_whisper
    >>> model = stable_whisper.load_model('base')
    >>> result = model.transcribe('audio.mp3')
    >>> result.save_as_json('audio.json')
    Saved: audio.json
    """
    if not isinstance(result, dict) and callable(getattr(result, 'to_dict')):
        result = result.to_dict()
    if not path.lower().endswith('.json'):
        path += '.json'
    result = json.dumps(result, allow_nan=True, ensure_ascii=ensure_ascii, **kwargs)
    _save_as_file(result, path)


def load_result(json_path: str) -> dict:
    """
    Return a ``dict`` of the contents in ``json_path``.
    """
    with open(json_path, 'r', encoding='utf-8') as f:
        return json.load(f)



import warnings
from .cli import cli
from .original_whisper import transcribe_stable, transcribe_minimal, load_model, modify_model
from .faster_whisper import load_faster_whisper
from .hf_whisper import load_hf_whisper


__all__ = ['load_model', 'modify_model', 'load_faster_whisper', 'load_hf_whisper']

warnings.filterwarnings('ignore', module='whisper', message='.*Triton.*', category=UserWarning)


if __name__ == '__main__':
    cli()



import os
from setuptools import setup, find_packages


def version() -> str:
    with open(os.path.join(os.path.dirname(__file__), 'stable_whisper/_version.py')) as f:
        return f.read().split('=')[-1].strip().strip('"').strip("'")


def read_me() -> str:
    with open('README.md', 'r', encoding='utf-8') as f:
        return f.read()


setup(
    name="stable-ts",
    version=version(),
    description="Modifies OpenAI's Whisper to produce more reliable timestamps.",
    long_description=read_me(),
    long_description_content_type='text/markdown',
    python_requires=">=3.8",
    author="Jian",
    url="https://github.com/jianfch/stable-ts",
    license="MIT",
    packages=find_packages(include=["stable_whisper", "stable_whisper.*"]),
    install_requires=[
        "numpy",
        "torch",
        "torchaudio",
        "tqdm",
        "openai-whisper==20231117"
    ],
    entry_points={
        "console_scripts": ["stable-ts=stable_whisper.whisper_word_level:cli"],
    },
    include_package_data=False
)
